import os
from typing import Any, Dict, Generator, Optional

import click
from click.testing import CliRunner
import pytest

from anyscale._private.models.image_uri import ANYSCALE_CLUSTER_ENV_PREFIX
from anyscale._private.sdk import _LAZY_SDK_SINGLETONS
from anyscale.commands.service_commands import deploy, status, wait
from anyscale.compute_config.models import ComputeConfig, HeadNodeConfig
from anyscale.service.commands import _SERVICE_SDK_SINGLETON_KEY
from anyscale.service.models import (
    ServiceConfig,
    ServiceState,
    ServiceStatus,
    ServiceVersionState,
    ServiceVersionStatus,
)


def _get_test_file_path(subpath: str) -> str:
    return os.path.join(os.path.dirname(__file__), "test_files/", subpath,)


EMPTY_CONFIG_PATH = _get_test_file_path("service_config_files/empty.yaml")
MINIMAL_CONFIG_PATH = _get_test_file_path("service_config_files/minimal.yaml")
FULL_CONFIG_PATH = _get_test_file_path("service_config_files/full.yaml")
NAME_AND_GIBBERISH_CONFIG_PATH = _get_test_file_path(
    "service_config_files/name_and_gibberish.yaml"
)
MULTI_LINE_REQUIREMENTS_PATH = _get_test_file_path("requirements_files/multi_line.txt")


class FakeServiceSDK:
    DEFAULT_SERVICE_ID = "default-fake-service-id"
    DEFAULT_SERVICE_NAME = "default-fake-service-name"

    def __init__(self):
        self.deployed_config: Optional[ServiceConfig] = None
        self.deployed_kwargs: Dict[str, Any] = {}
        self.fetched_name: Optional[str] = None
        self.waited_name: Optional[str] = None
        self.waited_state: Optional[ServiceState] = None
        self.waited_timeout_s: Optional[float] = None

    def deploy(self, config: ServiceConfig, **kwargs):
        assert isinstance(config, ServiceConfig)
        self.deployed_config = config
        self.deployed_kwargs = kwargs

    def status(
        self, name: str, *, cloud: Optional[str] = None, project: Optional[str] = None
    ) -> ServiceStatus:
        self.fetched_name = name
        return ServiceStatus(
            id=self.DEFAULT_SERVICE_ID,
            name=name or self.DEFAULT_SERVICE_NAME,
            state=ServiceState.TERMINATED,
            query_url="http://fake-service-url/",
            query_auth_token="asdf1234",
            primary_version=ServiceVersionStatus(
                id="primary",
                state=ServiceVersionState.RUNNING,
                weight=100,
                config=ServiceConfig(applications=[{"import_path": "main:app"}]),
            ),
            canary_version=ServiceVersionStatus(
                id="canary",
                state=ServiceVersionState.STARTING,
                weight=0,
                config=ServiceConfig(applications=[{"import_path": "main:app"}]),
            ),
        )

    def wait(
        self,
        name: str,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        state: ServiceState,
        timeout_s: float,
        interval_s: float,
    ):
        self.waited_name = name
        self.waited_state = state
        self.waited_timeout_s = timeout_s


@pytest.fixture()
def fake_service_sdk() -> Generator[FakeServiceSDK, None, None]:
    fake_service_sdk = FakeServiceSDK()
    _LAZY_SDK_SINGLETONS[_SERVICE_SDK_SINGLETON_KEY] = fake_service_sdk
    try:
        yield fake_service_sdk
    finally:
        del _LAZY_SDK_SINGLETONS[_SERVICE_SDK_SINGLETON_KEY]


def _assert_error_message(result: click.testing.Result, *, message: str):
    assert result.exit_code != 0
    assert message in result.stdout


class TestDeploy:
    def test_deploy_no_arg(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy)
        _assert_error_message(
            result, message="Either config file or import path must be provided."
        )

    def test_deploy_from_import_path(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

    def test_deploy_from_import_path_with_args(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "arg1=val1", "arg2=val2"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[
                {"import_path": "main:app", "args": {"arg1": "val1", "arg2": "val2"},}
            ],
        )
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

    def test_deploy_from_import_path_with_bad_arg(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "bad_arg"])
        _assert_error_message(
            result,
            message="Invalid key-value string 'bad_arg'. Must be of the form 'key=value'.",
        )

    def test_deploy_from_file(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

    def test_deploy_from_file_with_import_path(self, fake_service_sdk):
        runner = CliRunner()
        os.path.join(
            os.path.dirname(__file__), "test_service_config_files", "minimal.yaml",
        )

        result = runner.invoke(
            deploy, ["-f", MINIMAL_CONFIG_PATH, "main:app", "arg1=val1"]
        )
        _assert_error_message(
            result,
            message="When a config file is provided, import path and application arguments can't be.",
        )

        result = runner.invoke(deploy, ["-f", MINIMAL_CONFIG_PATH, "main:app"])
        _assert_error_message(
            result,
            message="When a config file is provided, import path and application arguments can't be.",
        )

    @pytest.mark.parametrize("flag", ["--in-place", "-i"])
    def test_deploy_in_place(self, fake_service_sdk, flag: str):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", flag])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )
        assert fake_service_sdk.deployed_kwargs == {
            "canary_percent": None,
            "in_place": True,
            "max_surge_percent": None,
        }

    def test_deploy_canary_percent(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "--canary-percent", "50"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )
        assert fake_service_sdk.deployed_kwargs == {
            "canary_percent": 50,
            "in_place": False,
            "max_surge_percent": None,
        }

    def test_deploy_max_surge_percent(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "--max-surge-percent", "50"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )
        assert fake_service_sdk.deployed_kwargs == {
            "canary_percent": None,
            "in_place": False,
            "max_surge_percent": 50,
        }

    def test_deploy_excludes(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

        # Pass a single exclusion.
        result = runner.invoke(deploy, ["--exclude", "path1", "main:app"])
        assert result.exit_code == 0, result.stdout
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}], excludes=["path1"]
        )
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

        # Pass multiple exclusions.
        result = runner.invoke(
            deploy, ["--exclude", "path1", "-e", "path2", "main:app"]
        )
        assert result.exit_code == 0, result.stdout
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}], excludes=["path1", "path2"]
        )
        assert (
            fake_service_sdk.deployed_config.applications
            == expected_config.applications
        )

    def test_deploy_both_image_uri_and_containerfile(self):
        runner = CliRunner()
        result = runner.invoke(
            deploy, ["main:app", "--image-uri", "image", "--containerfile", "file"]
        )
        _assert_error_message(
            result,
            message="Only one of '--containerfile' and '--image-uri' can be provided.",
        )

    def test_deploy_registry_login_secret_without_custom_image_uri(self,):
        runner = CliRunner()

        # No image uri provided
        result = runner.invoke(
            deploy, ["main:app", "--registry-login-secret", "secret",],
        )
        _assert_error_message(
            result,
            message="Registry login secret can only be used with an image that is not hosted on Anyscale.",
        )

        # Anyscale image uri provided
        result = runner.invoke(
            deploy,
            [
                "main:app",
                "--registry-login-secret",
                "secret",
                "--image-uri",
                f"{ANYSCALE_CLUSTER_ENV_PREFIX}some-image",
            ],
        )
        _assert_error_message(
            result,
            message="Registry login secret can only be used with an image that is not hosted on Anyscale.",
        )

    def test_deploy_from_file_override_options(self, fake_service_sdk):
        runner = CliRunner()

        # No overrides, should match the config in the file.
        result = runner.invoke(deploy, ["-f", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        inline_compute_config = ComputeConfig(
            cloud="test-cloud", head_node=HeadNodeConfig(instance_type="m5.2xlarge")
        )

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config=inline_compute_config,
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"]},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override name.
        result = runner.invoke(
            deploy, ["-f", FULL_CONFIG_PATH, "--name", "override-name"]
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="override-name",
            image_uri="docker.io/library/test:latest",
            compute_config=inline_compute_config,
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"]},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override image URI.
        result = runner.invoke(
            deploy,
            [
                "-f",
                FULL_CONFIG_PATH,
                "--image-uri",
                "docker.io/user/override-image:latest",
                "--registry-login-secret",
                "secret-name",
            ],
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/user/override-image:latest",
            registry_login_secret="secret-name",
            compute_config=inline_compute_config,
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"]},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override compute_config.
        result = runner.invoke(
            deploy,
            ["-f", FULL_CONFIG_PATH, "--compute-config", "override-compute-config"],
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config="override-compute-config",
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"]},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override working_dir.
        result = runner.invoke(
            deploy, ["-f", FULL_CONFIG_PATH, "--working-dir", "override-working-dir"]
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config=inline_compute_config,
            working_dir="override-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"]},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override requirements.
        result = runner.invoke(
            deploy,
            ["-f", FULL_CONFIG_PATH, "--requirements", MULTI_LINE_REQUIREMENTS_PATH],
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config=inline_compute_config,
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=MULTI_LINE_REQUIREMENTS_PATH,
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"]},
            logging_config={"log_level": "DEBUG"},
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override environment variables.
        result = runner.invoke(
            deploy,
            [
                "-f",
                FULL_CONFIG_PATH,
                "--env",
                "FOO=BAR_override",
                "--env",
                "FOO2=BAR2",
            ],
        )
        print(result.stdout)
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None
        expected_config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config=inline_compute_config,
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"]},
            logging_config={"log_level": "DEBUG"},
            env_vars={"FOO": "BAR_override", "FOO2": "BAR2"},
        )
        assert fake_service_sdk.deployed_config == expected_config


class TestStatus:
    def test_no_name(self, fake_service_sdk):
        # Test passing no name (would be picked up as workspace name).
        runner = CliRunner()
        result = runner.invoke(status)
        _assert_error_message(
            result,
            message="Service name must be provided using '--name' or in a config file using '-f'",
        )

    def test_name_and_config_file(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-n", "custom-name", "-f", "file-name.yaml"])
        _assert_error_message(
            result, message="Only one of '--name' and '--config-file' can be provided."
        )

    def test_name(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-n", "custom-name"])
        assert result.exit_code == 0
        assert fake_service_sdk.fetched_name == "custom-name"

    def test_service_config_file_not_found(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-f", "nonexistent.yaml"])
        _assert_error_message(
            result, message="Config file not found at path: 'nonexistent.yaml'."
        )

    def test_empty_config_file(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-f", EMPTY_CONFIG_PATH])
        _assert_error_message(
            result,
            message=f"No 'name' property found in config file '{EMPTY_CONFIG_PATH}'.",
        )

    def test_service_config_file_missing_name(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-f", MINIMAL_CONFIG_PATH])
        _assert_error_message(
            result,
            message=f"No 'name' property found in config file '{MINIMAL_CONFIG_PATH}'.",
        )

    def test_service_config_file(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-f", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.fetched_name == "test-name-from-file"

    def test_service_config_file_only_reads_name(self, fake_service_sdk):
        # `status` CLI path should ignore any other fields in the config file aside from `name`.
        runner = CliRunner()
        result = runner.invoke(status, ["-f", NAME_AND_GIBBERISH_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.fetched_name == "service-name-from-file"

    def test_verbose_flag(self, fake_service_sdk):
        runner = CliRunner()

        # No verbose flag -- exclude details.
        result = runner.invoke(status, ["-n", "custom-name"])
        assert result.exit_code == 0
        assert fake_service_sdk.fetched_name == "custom-name"
        assert "primary_version" in result.stdout
        assert "canary_version" in result.stdout
        assert "query_url" in result.stdout
        assert "query_auth_token" in result.stdout
        assert "config" not in result.stdout
        assert "applications" not in result.stdout

        # Verbose flag -- include details.
        result = runner.invoke(status, ["-n", "custom-name", "-v"])
        assert result.exit_code == 0
        assert fake_service_sdk.fetched_name == "custom-name"
        assert "primary_version" in result.stdout
        assert "canary_version" in result.stdout
        assert "query_url" in result.stdout
        assert "query_auth_token" in result.stdout
        assert "config" in result.stdout
        assert "applications" in result.stdout


class TestWait:
    def test_no_name(self, fake_service_sdk):
        # Test passing no name (would be picked up as workspace name).
        runner = CliRunner()
        result = runner.invoke(wait)
        _assert_error_message(
            result,
            message="Service name must be provided using '--name' or in a config file using '-f'",
        )

    def test_name(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-n", "custom-name"])
        assert result.exit_code == 0
        assert fake_service_sdk.waited_name == "custom-name"

    def test_service_config_file_not_found(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-f", "nonexistent.yaml"])
        _assert_error_message(
            result, message="Config file not found at path: 'nonexistent.yaml'."
        )

    def test_empty_config_file(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-f", EMPTY_CONFIG_PATH])
        _assert_error_message(
            result,
            message=f"No 'name' property found in config file '{EMPTY_CONFIG_PATH}'.",
        )

    def test_service_config_file_missing_name(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-f", MINIMAL_CONFIG_PATH])
        _assert_error_message(
            result,
            message=f"No 'name' property found in config file '{MINIMAL_CONFIG_PATH}'.",
        )

    def test_service_config_file(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-f", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.waited_name == "test-name-from-file"

    def test_service_config_file_only_reads_name(self, fake_service_sdk):
        # `wait` CLI path should ignore any other fields in the config file aside from `name`.
        runner = CliRunner()
        result = runner.invoke(wait, ["-f", NAME_AND_GIBBERISH_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.waited_name == "service-name-from-file"

    def test_state(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-n", "custom-name", "--state", "TERMINATED"])
        assert result.exit_code == 0
        assert fake_service_sdk.waited_name == "custom-name"
        assert fake_service_sdk.waited_state == ServiceState.TERMINATED

    def test_bad_state(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-n", "custom-name", "--state", "FOOBAR"])
        assert result.exit_code == 1
        assert "'FOOBAR' is not a valid ServiceState" in str(result.stdout)

    def test_timeout(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-n", "custom-name", "--timeout-s=30"])
        assert result.exit_code == 0
        assert fake_service_sdk.waited_timeout_s == 30



import os
from typing import cast, List, Optional

from anyscale._private.anyscale_client.common import AnyscaleClientInterface
from anyscale._private.models.image_uri import ImageURI
from anyscale._private.sdk.base_sdk import BaseSDK
from anyscale.cli_logger import BlockLogger
from anyscale.image.models import ImageBuild, ImageBuildStatus
from anyscale.sdk.anyscale_client import (
    ClusterEnvironmentBuild,
    ClusterEnvironmentBuildStatus,
)


ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS = (
    os.environ.get("ANYSCALE_ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS", "0") == "1"
)


class ImageSDK(BaseSDK):
    def __init__(
        self,
        *,
        logger: Optional[BlockLogger] = None,
        client: Optional[AnyscaleClientInterface] = None,
    ):
        super().__init__(logger=logger, client=client)
        self._enable_image_build_for_tracked_requirements = (
            ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS
        )

    @property
    def enable_image_build_for_tracked_requirements(self) -> bool:
        return self._enable_image_build_for_tracked_requirements

    def get_default_image(self) -> str:
        return self.client.get_default_build_id()

    def get_image_build(self, build_id: str) -> Optional[ClusterEnvironmentBuild]:
        return self.client.get_cluster_env_build(build_id)

    _BACKEND_IMAGE_STATUS_TO_IMAGE_BUILD_STATUS = {
        ClusterEnvironmentBuildStatus.PENDING: ImageBuildStatus.IN_PROGRESS,
        ClusterEnvironmentBuildStatus.IN_PROGRESS: ImageBuildStatus.IN_PROGRESS,
        ClusterEnvironmentBuildStatus.SUCCEEDED: ImageBuildStatus.SUCCEEDED,
        ClusterEnvironmentBuildStatus.FAILED: ImageBuildStatus.FAILED,
        ClusterEnvironmentBuildStatus.PENDING_CANCELLATION: ImageBuildStatus.FAILED,
        ClusterEnvironmentBuildStatus.CANCELED: ImageBuildStatus.FAILED,
    }

    def _get_image_build_status(
        self, build: ClusterEnvironmentBuild
    ) -> ImageBuildStatus:
        return cast(
            ImageBuildStatus,
            self._BACKEND_IMAGE_STATUS_TO_IMAGE_BUILD_STATUS.get(
                build.status, ImageBuildStatus.UNKNOWN
            ),
        )

    def get(self, name: str) -> ImageBuild:
        if ":" in name:
            name, version = name.split(":", 2)
        else:
            version = None
        cluster_env = self.client.get_cluster_env_by_name(name=name)
        if cluster_env:
            build = None
            for b in self.client.list_cluster_env_builds(cluster_env.id):
                if version is None:
                    # use the latest
                    build = b
                    break
                if b.revision == int(version):
                    build = b
                    break

            if not build:
                raise RuntimeError(f"Version {version} not found for image {name}.")

            image_uri = ImageURI.from_cluster_env_build(
                cluster_env=cluster_env, build=build
            )
            return ImageBuild(
                status=self._get_image_build_status(build),
                uri=str(image_uri),
                ray_version=build.ray_version,
            )
        else:
            raise RuntimeError(f"Image {name} not found.")

    def build_image_from_containerfile(
        self,
        name: str,
        containerfile: str,
        ray_version: Optional[str] = None,
        anonymous: bool = False,
    ) -> str:
        return self.client.get_cluster_env_build_id_from_containerfile(
            cluster_env_name=name,
            containerfile=containerfile,
            anonymous=anonymous,
            ray_version=ray_version,
        )

    def build_image_from_requirements(
        self, name: str, base_build_id: str, requirements: List[str]
    ):
        if requirements:
            base_build = self.client.get_cluster_env_build(base_build_id)
            if (
                base_build
                and base_build.status == ClusterEnvironmentBuildStatus.SUCCEEDED
                and base_build.docker_image_name
            ):
                self.logger.info(f"Using tracked python packages: {requirements}")
                lines = [
                    "# syntax=docker/dockerfile:1",
                    f"FROM {base_build.docker_image_name}",
                ]
                for requirement in requirements:
                    lines.append(f'RUN pip install "{requirement}"')
                return self.build_image_from_containerfile(
                    name, "\n".join(lines), ray_version=base_build.ray_version
                )
            else:
                raise RuntimeError(
                    f"Base build {base_build_id} is not a successful build."
                )
        else:
            return base_build_id

    def registery_image(
        self,
        image_uri: str,
        registry_login_secret: Optional[str] = None,
        ray_version: Optional[str] = None,
    ) -> str:
        image_uri_checked = ImageURI.from_str(image_uri_str=image_uri)
        return self.client.get_cluster_env_build_id_from_image_uri(
            image_uri=image_uri_checked,
            registry_login_secret=registry_login_secret,
            ray_version=ray_version,
        )

    def get_image_uri_from_build_id(self, build_id: str) -> Optional[ImageURI]:
        return self.client.get_cluster_env_build_image_uri(
            cluster_env_build_id=build_id
        )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class BufferRegistration(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'aws': 'NodeRegistrationAWS',
        'gcp': 'NodeRegistrationGCP',
        'provisioned': 'NodeRegistrationProvisioned',
        'sagemaker': 'NodeRegistrationSagemaker',
        'provider': 'CloudProvider',
        'cloud_id': 'str',
        'buffer_id': 'str'
    }

    attribute_map = {
        'aws': 'aws',
        'gcp': 'gcp',
        'provisioned': 'provisioned',
        'sagemaker': 'sagemaker',
        'provider': 'provider',
        'cloud_id': 'cloud_id',
        'buffer_id': 'buffer_id'
    }

    def __init__(self, aws=None, gcp=None, provisioned=None, sagemaker=None, provider=None, cloud_id=None, buffer_id=None, local_vars_configuration=None):  # noqa: E501
        """BufferRegistration - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._aws = None
        self._gcp = None
        self._provisioned = None
        self._sagemaker = None
        self._provider = None
        self._cloud_id = None
        self._buffer_id = None
        self.discriminator = None

        if aws is not None:
            self.aws = aws
        if gcp is not None:
            self.gcp = gcp
        if provisioned is not None:
            self.provisioned = provisioned
        if sagemaker is not None:
            self.sagemaker = sagemaker
        self.provider = provider
        self.cloud_id = cloud_id
        self.buffer_id = buffer_id

    @property
    def aws(self):
        """Gets the aws of this BufferRegistration.  # noqa: E501


        :return: The aws of this BufferRegistration.  # noqa: E501
        :rtype: NodeRegistrationAWS
        """
        return self._aws

    @aws.setter
    def aws(self, aws):
        """Sets the aws of this BufferRegistration.


        :param aws: The aws of this BufferRegistration.  # noqa: E501
        :type: NodeRegistrationAWS
        """

        self._aws = aws

    @property
    def gcp(self):
        """Gets the gcp of this BufferRegistration.  # noqa: E501


        :return: The gcp of this BufferRegistration.  # noqa: E501
        :rtype: NodeRegistrationGCP
        """
        return self._gcp

    @gcp.setter
    def gcp(self, gcp):
        """Sets the gcp of this BufferRegistration.


        :param gcp: The gcp of this BufferRegistration.  # noqa: E501
        :type: NodeRegistrationGCP
        """

        self._gcp = gcp

    @property
    def provisioned(self):
        """Gets the provisioned of this BufferRegistration.  # noqa: E501


        :return: The provisioned of this BufferRegistration.  # noqa: E501
        :rtype: NodeRegistrationProvisioned
        """
        return self._provisioned

    @provisioned.setter
    def provisioned(self, provisioned):
        """Sets the provisioned of this BufferRegistration.


        :param provisioned: The provisioned of this BufferRegistration.  # noqa: E501
        :type: NodeRegistrationProvisioned
        """

        self._provisioned = provisioned

    @property
    def sagemaker(self):
        """Gets the sagemaker of this BufferRegistration.  # noqa: E501


        :return: The sagemaker of this BufferRegistration.  # noqa: E501
        :rtype: NodeRegistrationSagemaker
        """
        return self._sagemaker

    @sagemaker.setter
    def sagemaker(self, sagemaker):
        """Sets the sagemaker of this BufferRegistration.


        :param sagemaker: The sagemaker of this BufferRegistration.  # noqa: E501
        :type: NodeRegistrationSagemaker
        """

        self._sagemaker = sagemaker

    @property
    def provider(self):
        """Gets the provider of this BufferRegistration.  # noqa: E501

        Which provider this registration is for.  # noqa: E501

        :return: The provider of this BufferRegistration.  # noqa: E501
        :rtype: CloudProvider
        """
        return self._provider

    @provider.setter
    def provider(self, provider):
        """Sets the provider of this BufferRegistration.

        Which provider this registration is for.  # noqa: E501

        :param provider: The provider of this BufferRegistration.  # noqa: E501
        :type: CloudProvider
        """
        if self.local_vars_configuration.client_side_validation and provider is None:  # noqa: E501
            raise ValueError("Invalid value for `provider`, must not be `None`")  # noqa: E501

        self._provider = provider

    @property
    def cloud_id(self):
        """Gets the cloud_id of this BufferRegistration.  # noqa: E501

        Cloud ID of the buffer node.  # noqa: E501

        :return: The cloud_id of this BufferRegistration.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this BufferRegistration.

        Cloud ID of the buffer node.  # noqa: E501

        :param cloud_id: The cloud_id of this BufferRegistration.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def buffer_id(self):
        """Gets the buffer_id of this BufferRegistration.  # noqa: E501

        Buffer ID of the buffer node.  # noqa: E501

        :return: The buffer_id of this BufferRegistration.  # noqa: E501
        :rtype: str
        """
        return self._buffer_id

    @buffer_id.setter
    def buffer_id(self, buffer_id):
        """Sets the buffer_id of this BufferRegistration.

        Buffer ID of the buffer node.  # noqa: E501

        :param buffer_id: The buffer_id of this BufferRegistration.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and buffer_id is None:  # noqa: E501
            raise ValueError("Invalid value for `buffer_id`, must not be `None`")  # noqa: E501

        self._buffer_id = buffer_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, BufferRegistration):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, BufferRegistration):
            return True

        return self.to_dict() != other.to_dict()



from typing import Any, cast, Dict, List, Optional, Union
import uuid

from anyscale._private.workload import WorkloadSDK
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import (
    CreateInternalProductionJob,
    InternalProductionJob,
    ProductionJobConfig,
)
from anyscale.client.openapi_client.models.job_status import (
    JobStatus as BackendJobStatus,
)
from anyscale.client.openapi_client.models.production_job import ProductionJob
from anyscale.client.openapi_client.models.ray_runtime_env_config import (
    RayRuntimeEnvConfig,
)
from anyscale.compute_config.models import ComputeConfig
from anyscale.job.models import (
    JobConfig,
    JobLogMode,
    JobRunState,
    JobRunStatus,
    JobState,
    JobStatus,
)
from anyscale.sdk.anyscale_client.models import Job
from anyscale.sdk.anyscale_client.models.ha_job_states import HaJobStates
from anyscale.utils.runtime_env import parse_requirements_file


logger = BlockLogger()


class PrivateJobSDK(WorkloadSDK):
    _POLLING_INTERVAL_SECONDS = 10.0

    def _populate_runtime_env(
        self,
        config: JobConfig,
        *,
        autopopulate_in_workspace: bool = True,
        cloud_id: Optional[str] = None,
        workspace_requirements_path: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """Populates a runtime_env from the config.

        Local directories specified in the 'working_dir' will be uploaded and
        replaced with the resulting remote URIs.

        Requirements files will be loaded and populated into the 'pip' field.

        If autopopulate_from_workspace is passed and this code is running inside a
        workspace, the following defaults will be applied:
            - 'working_dir' will be set to '.'.
            - 'pip' will be set to the workspace-managed requirements file.
        """
        runtime_env: Dict[str, Any] = {}
        [runtime_env] = self.override_and_upload_local_dirs(
            [runtime_env],
            working_dir_override=config.working_dir,
            excludes_override=config.excludes,
            cloud_id=cloud_id,
            autopopulate_in_workspace=autopopulate_in_workspace,
            additional_py_modules=config.py_modules,
        )
        [runtime_env] = self.override_and_load_requirements_files(
            [runtime_env],
            requirements_override=config.requirements,
            workspace_requirements_path=workspace_requirements_path,
        )
        [runtime_env] = self.update_env_vars(
            [runtime_env], env_vars_updates=config.env_vars,
        )

        return runtime_env or None

    def _get_default_name(self) -> str:
        """Get a default name for the job.

        If running inside a workspace, this is generated from the workspace name,
        else it generates a random name.
        """
        # TODO(edoakes): generate two random words instead of UUID here.
        name = f"job-{self.get_current_workspace_name() or str(uuid.uuid4())}"
        self.logger.info(f"No name was specified, using default: '{name}'.")
        return name

    def submit(self, config: JobConfig) -> str:
        name = config.name or self._get_default_name()

        build_id = None
        if config.containerfile is not None:
            build_id = self._image_sdk.build_image_from_containerfile(
                name=f"image-for-job-{name}",
                containerfile=self.get_containerfile_contents(config.containerfile),
                ray_version=config.ray_version,
            )
        elif config.image_uri is not None:
            build_id = self._image_sdk.registery_image(
                image_uri=config.image_uri,
                registry_login_secret=config.registry_login_secret,
                ray_version=config.ray_version,
            )

        if self._image_sdk.enable_image_build_for_tracked_requirements:
            requirements_path_to_be_populated_in_runtime_env = None
            requirements_path = self.client.get_workspace_requirements_path()
            if requirements_path is not None:
                requirements = parse_requirements_file(requirements_path)
                if requirements:
                    build_id = self._image_sdk.build_image_from_requirements(
                        name=f"image-for-job-{name}",
                        base_build_id=self.client.get_default_build_id(),
                        requirements=requirements,
                    )
        else:
            requirements_path_to_be_populated_in_runtime_env = (
                self.client.get_workspace_requirements_path()
            )

        if build_id is None:
            build_id = self.client.get_default_build_id()

        env_vars_from_workspace = self.client.get_workspace_env_vars()
        if env_vars_from_workspace:
            if config.env_vars:
                # the precedence should be cli > workspace
                env_vars_from_workspace.update(config.env_vars)
                config = config.options(env_vars=env_vars_from_workspace)
            else:
                config = config.options(env_vars=env_vars_from_workspace)

        compute_config_id, cloud_id = self.resolve_compute_config_and_cloud_id(
            compute_config=config.compute_config, cloud=config.cloud
        )

        project_id = self.client.get_project_id(
            parent_cloud_id=cloud_id, name=config.project
        )

        runtime_env = self._populate_runtime_env(
            config,
            cloud_id=cloud_id,
            workspace_requirements_path=requirements_path_to_be_populated_in_runtime_env,
        )

        job: InternalProductionJob = self.client.submit_job(
            CreateInternalProductionJob(
                name=name,
                project_id=project_id,
                workspace_id=self.client.get_current_workspace_id(),
                config=ProductionJobConfig(
                    entrypoint=config.entrypoint,
                    runtime_env=runtime_env,
                    build_id=build_id,
                    compute_config_id=compute_config_id,
                    max_retries=config.max_retries,
                ),
                # TODO(edoakes): support job queue config.
                job_queue_config=None,
            )
        )

        self.logger.info(f"Job '{name}' submitted, ID: '{job.id}'.")
        self.logger.info(
            f"View the job in the UI: {self.client.get_job_ui_url(job.id)}"
        )
        return job.id

    _HA_JOB_STATE_TO_JOB_STATE = {
        HaJobStates.UPDATING: JobState.RUNNING,
        HaJobStates.RUNNING: JobState.RUNNING,
        HaJobStates.RESTARTING: JobState.RUNNING,
        HaJobStates.CLEANING_UP: JobState.RUNNING,
        HaJobStates.PENDING: JobState.STARTING,
        HaJobStates.AWAITING_CLUSTER_START: JobState.STARTING,
        HaJobStates.SUCCESS: JobState.SUCCEEDED,
        HaJobStates.ERRORED: JobState.FAILED,
        HaJobStates.TERMINATED: JobState.FAILED,
        HaJobStates.BROKEN: JobState.FAILED,
        HaJobStates.OUT_OF_RETRIES: JobState.FAILED,
    }

    _BACKEND_JOB_STATUS_TO_JOB_RUN_STATE = {
        BackendJobStatus.RUNNING: JobRunState.RUNNING,
        BackendJobStatus.COMPLETED: JobRunState.SUCCEEDED,
        BackendJobStatus.PENDING: JobRunState.STARTING,
        BackendJobStatus.STOPPED: JobRunState.FAILED,
        BackendJobStatus.SUCCEEDED: JobRunState.SUCCEEDED,
        BackendJobStatus.FAILED: JobRunState.FAILED,
        BackendJobStatus.UNKNOWN: JobRunState.UNKNOWN,
    }

    def _job_state_from_job_model(self, model: ProductionJob) -> JobState:
        ha_state = model.state.current_state if model.state else None
        return cast(
            JobState, self._HA_JOB_STATE_TO_JOB_STATE.get(ha_state, JobState.UNKNOWN)
        )

    def _job_run_model_to_job_run_status(self, run: Job) -> JobRunStatus:
        state = self._BACKEND_JOB_STATUS_TO_JOB_RUN_STATE.get(
            run.status, JobRunState.UNKNOWN
        )
        return JobRunStatus(name=run.name, state=state)

    def _job_model_to_status(self, model: ProductionJob, runs: List[Job]) -> JobStatus:
        state = self._job_state_from_job_model(model)
        project = self.client.get_project(model.project_id)

        prod_job_config: ProductionJobConfig = model.config
        runtime_env_config: RayRuntimeEnvConfig = prod_job_config.runtime_env if prod_job_config else None
        compute_config = self.get_user_facing_compute_config(
            prod_job_config.compute_config_id
        )
        config = JobConfig(
            name=model.name,
            compute_config=compute_config,
            requirements=runtime_env_config.pip if runtime_env_config else None,
            working_dir=runtime_env_config.working_dir if runtime_env_config else None,
            env_vars=runtime_env_config.env_vars if runtime_env_config else None,
            entrypoint=prod_job_config.entrypoint,
            cloud=compute_config.cloud
            if compute_config and isinstance(compute_config, ComputeConfig)
            else None,
            max_retries=prod_job_config.max_retries
            if prod_job_config.max_retries is not None
            else -1,
            project=project.name
            if project is not None and project.name != "default"
            else None,
        )
        runs = [self._job_run_model_to_job_run_status(run) for run in runs]

        return JobStatus(
            name=model.name, id=model.id, state=state, runs=runs, config=config
        )

    def _resolve_to_job_model(
        self,
        *,
        name: Optional[str] = None,
        job_id: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> ProductionJob:
        if name is None and job_id is None:
            raise ValueError("One of 'name' or 'job_id' must be provided.")

        if name is not None and job_id is not None:
            raise ValueError("Only one of 'name' or 'job_id' can be provided.")

        if job_id is not None and (cloud is not None or project is not None):
            raise ValueError("'cloud' and 'project' should only be used with 'name'.")

        model: Optional[ProductionJob] = self.client.get_job(
            name=name, job_id=job_id, cloud=cloud, project=project
        )
        if model is None:
            if name is not None:
                raise RuntimeError(f"Job with name '{name}' was not found.")
            else:
                raise RuntimeError(f"Job with ID '{job_id}' was not found.")

        return model

    def status(
        self,
        *,
        name: Optional[str] = None,
        job_id: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> JobStatus:
        job_model = self._resolve_to_job_model(
            name=name, job_id=job_id, cloud=cloud, project=project
        )
        runs = self.client.get_job_runs(job_model.id)
        return self._job_model_to_status(model=job_model, runs=runs)

    def terminate(
        self,
        *,
        job_id: Optional[str] = None,
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        job_model = self._resolve_to_job_model(
            name=name, job_id=job_id, cloud=cloud, project=project
        )
        self.client.terminate_job(job_model.id)
        self.logger.info(f"Marked job '{job_model.name}' for termination")
        return job_model.id

    _TERMINAL_HA_JOB_STATES = [
        HaJobStates.SUCCESS,
        HaJobStates.TERMINATED,
        HaJobStates.OUT_OF_RETRIES,
    ]

    def archive(
        self,
        *,
        job_id: Optional[str] = None,
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        job_model = self._resolve_to_job_model(
            name=name, job_id=job_id, cloud=cloud, project=project
        )

        ha_state = job_model.state.current_state if job_model.state else None
        if ha_state not in self._TERMINAL_HA_JOB_STATES:
            raise RuntimeError(
                f"Job with ID '{job_model.id}' has not reached a terminal state and cannot be archived."
            )

        self.client.archive_job(job_model.id)
        self.logger.info(f"Job {job_model.id} is successfully archived.")
        return job_model.id

    def wait(
        self,
        *,
        name: Optional[str] = None,
        job_id: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        state: Union[str, JobState] = JobState.SUCCEEDED,
        timeout_s: float = 1800,
        interval_s: float = _POLLING_INTERVAL_SECONDS,
    ):
        if not isinstance(timeout_s, (int, float)):
            raise TypeError("timeout_s must be a float")
        if timeout_s <= 0:
            raise ValueError("timeout_s must be >= 0")

        if not isinstance(interval_s, (int, float)):
            raise TypeError("interval_s must be a float")
        if interval_s <= 0:
            raise ValueError("interval_s must be >= 0")

        if not isinstance(state, JobState):
            raise TypeError("'state' must be a JobState.")

        job_id_or_name = job_id or name
        job_model = self._resolve_to_job_model(
            name=name, job_id=job_id, cloud=cloud, project=project
        )
        curr_state = self._job_state_from_job_model(job_model)

        self.logger.info(
            f"Waiting for job '{job_id_or_name}' to reach target state {state}, currently in state: {curr_state}"
        )
        for _ in self.timer.poll(timeout_s=timeout_s, interval_s=interval_s):
            job_model = self._resolve_to_job_model(name=name, job_id=job_id)
            new_state = self._job_state_from_job_model(job_model)

            if new_state != curr_state:
                self.logger.info(
                    f"Job '{job_id_or_name}' transitioned from {curr_state} to {new_state}"
                )
                curr_state = new_state

            if curr_state == state:
                self.logger.info(
                    f"Job '{job_id_or_name}' reached target state, exiting"
                )
                break

            if JobState.is_terminal(curr_state):
                raise RuntimeError(
                    f"Job '{job_id_or_name}' reached terminal state '{curr_state}', and will not reach '{state}'."
                )
        else:
            raise TimeoutError(
                f"Job '{job_id_or_name}' did not reach target state {state} within {timeout_s}s. Last seen state: {curr_state}."
            )

    def _resolve_job_run_id(
        self,
        *,
        job_id: Optional[str] = None,
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        run: Optional[str] = None,
    ) -> str:
        job_model = self._resolve_to_job_model(
            name=name, job_id=job_id, cloud=cloud, project=project
        )

        last_job_run_id = job_model.last_job_run_id
        if last_job_run_id is None:
            return ""
        if run is None:
            job_run_id = last_job_run_id
        else:
            runs: List[Job] = self.client.get_job_runs(job_model.id)
            for job_run in runs:
                if job_run.name == run:
                    job_run_id = job_run.id
                    break
            else:
                raise ValueError(
                    f"Job run '{run}' was not found for job '{job_id or name}'."
                )

        return job_run_id

    def get_logs(
        self,
        *,
        job_id: Optional[str] = None,
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        run: Optional[str] = None,
        mode: Union[str, JobLogMode] = JobLogMode.TAIL,
        max_lines: Optional[int] = None,
    ) -> str:
        if max_lines is not None:
            if not isinstance(max_lines, int):
                raise TypeError("max_lines must be an int")
            if max_lines <= 0:
                raise ValueError("max_lines must be > 0")

        job_run_id = self._resolve_job_run_id(
            job_id=job_id, name=name, cloud=cloud, project=project, run=run
        )

        head = mode == JobLogMode.HEAD
        return self.client.logs_for_job_run(
            job_run_id=job_run_id, head=head, max_lines=max_lines
        )



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
from enum import Enum


"""
Headers used within Anyscale
"""


class RequestHeaders(str, Enum):
    """
    Request headers, set by the client.
    """

    # Standardized authorization header
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Authorization
    AUTHORIZATION = "Authorization"

    # Identifies type of client (Web, SDK, CLI, etc.)
    CLIENT = "X-Anyscale-Client"

    # Identifies the version of the client
    CLIENT_VERSION = "X-Anyscale-Client-Version"


class ResponseHeaders(str, Enum):
    """
    Response headers, set by our server.
    """

    # Identifies a tenant for observability, used for authenticating requests to Cortex
    TENANT_ID = "X-Anyscale-Tenant-Id"

    # Identifies a user in Grafana, used for authenticating user requests to Grafana
    GRAFANA_USER_ID = "X-Anyscale-Grafana-User-Id"



# coding: utf-8

# flake8: noqa
"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

# import models into model package
from openapi_client.models.aws_credentials import AWSCredentials
from openapi_client.models.aws_memory_db_cluster_config import AWSMemoryDBClusterConfig
from openapi_client.models.aws_node_options import AWSNodeOptions
from openapi_client.models.aws_tag import AWSTag
from openapi_client.models.aws_tag_specification import AWSTagSpecification
from openapi_client.models.access_config import AccessConfig
from openapi_client.models.access_mode import AccessMode
from openapi_client.models.aica_endpoint import AicaEndpoint
from openapi_client.models.aica_endpoint_event import AicaEndpointEvent
from openapi_client.models.aica_endpoint_event_level import AicaEndpointEventLevel
from openapi_client.models.aica_endpoint_event_type import AicaEndpointEventType
from openapi_client.models.aica_endpoint_scope import AicaEndpointScope
from openapi_client.models.aica_model import AicaModel
from openapi_client.models.aica_model_accelerator_map import AicaModelAcceleratorMap
from openapi_client.models.aica_model_configuration import AicaModelConfiguration
from openapi_client.models.aica_observability_urls import AicaObservabilityUrls
from openapi_client.models.aicaendpoint_list_response import AicaendpointListResponse
from openapi_client.models.aicaendpoint_response import AicaendpointResponse
from openapi_client.models.aicaendpointevent_list_response import AicaendpointeventListResponse
from openapi_client.models.aicamodel_list_response import AicamodelListResponse
from openapi_client.models.aicamodel_response import AicamodelResponse
from openapi_client.models.aioa_cloud_waitlist_record import AioaCloudWaitlistRecord
from openapi_client.models.aioacloudwaitlistrecord_response import AioacloudwaitlistrecordResponse
from openapi_client.models.alert_type import AlertType
from openapi_client.models.anyscale_aws_account import AnyscaleAWSAccount
from openapi_client.models.anyscale_service_account import AnyscaleServiceAccount
from openapi_client.models.anyscale_version_response import AnyscaleVersionResponse
from openapi_client.models.anyscaleawsaccount_response import AnyscaleawsaccountResponse
from openapi_client.models.anyscaled_credential_response import AnyscaledCredentialResponse
from openapi_client.models.anyscaledcredentialresponse_response import AnyscaledcredentialresponseResponse
from openapi_client.models.anyscaleserviceaccount_response import AnyscaleserviceaccountResponse
from openapi_client.models.anyscaleversionresponse_response import AnyscaleversionresponseResponse
from openapi_client.models.api_key_parameters import ApiKeyParameters
from openapi_client.models.app_config import AppConfig
from openapi_client.models.app_config_config_schema import AppConfigConfigSchema
from openapi_client.models.appconfig_list_response import AppconfigListResponse
from openapi_client.models.appconfig_response import AppconfigResponse
from openapi_client.models.applied_snapshot import AppliedSnapshot
from openapi_client.models.apply_production_service_v2_model import ApplyProductionServiceV2Model
from openapi_client.models.archive_status import ArchiveStatus
from openapi_client.models.archived_logs_info import ArchivedLogsInfo
from openapi_client.models.archivedlogsinfo_response import ArchivedlogsinfoResponse
from openapi_client.models.attach_machine_pool_to_cloud_request import AttachMachinePoolToCloudRequest
from openapi_client.models.attachmachinepooltocloudresponse_response import AttachmachinepooltocloudresponseResponse
from openapi_client.models.autoscaler_credentials import AutoscalerCredentials
from openapi_client.models.autoscalercredentials_response import AutoscalercredentialsResponse
from openapi_client.models.aviary_model_config_v2 import AviaryModelConfigV2
from openapi_client.models.aws_region_and_zones import AwsRegionAndZones
from openapi_client.models.aws_region_info import AwsRegionInfo
from openapi_client.models.awsregionandzones_response import AwsregionandzonesResponse
from openapi_client.models.baseimagesenum import BASEIMAGESENUM
from openapi_client.models.bank_account_information import BankAccountInformation
from openapi_client.models.base_job_status import BaseJobStatus
from openapi_client.models.batch_response_batched_result_organization_invitation_base import BatchResponseBatchedResultOrganizationInvitationBase
from openapi_client.models.batched_result_organization_invitation_base import BatchedResultOrganizationInvitationBase
from openapi_client.models.billing_information import BillingInformation
from openapi_client.models.billing_version_code import BillingVersionCode
from openapi_client.models.block_device_mapping import BlockDeviceMapping
from openapi_client.models.body_upload_file_api_v2_files_post import BodyUploadFileApiV2FilesPost
from openapi_client.models.buffer_registration import BufferRegistration
from openapi_client.models.build import Build
from openapi_client.models.build_log_response import BuildLogResponse
from openapi_client.models.build_registration import BuildRegistration
from openapi_client.models.build_response import BuildResponse
from openapi_client.models.build_status import BuildStatus
from openapi_client.models.buildlogresponse_response import BuildlogresponseResponse
from openapi_client.models.card import Card
from openapi_client.models.card_id import CardId
from openapi_client.models.card_list_response import CardListResponse
from openapi_client.models.change_password_params import ChangePasswordParams
from openapi_client.models.cleanup_leaked_grafana_dashboard_response import CleanupLeakedGrafanaDashboardResponse
from openapi_client.models.cleanupleakedgrafanadashboardresponse_response import CleanupleakedgrafanadashboardresponseResponse
from openapi_client.models.clone_experimental_workspace import CloneExperimentalWorkspace
from openapi_client.models.cloud import Cloud
from openapi_client.models.cloud_analytics_event import CloudAnalyticsEvent
from openapi_client.models.cloud_analytics_event_cloud_provider_error import CloudAnalyticsEventCloudProviderError
from openapi_client.models.cloud_analytics_event_cloud_resource import CloudAnalyticsEventCloudResource
from openapi_client.models.cloud_analytics_event_command_name import CloudAnalyticsEventCommandName
from openapi_client.models.cloud_analytics_event_error import CloudAnalyticsEventError
from openapi_client.models.cloud_analytics_event_name import CloudAnalyticsEventName
from openapi_client.models.cloud_collaborator import CloudCollaborator
from openapi_client.models.cloud_collaborator_value import CloudCollaboratorValue
from openapi_client.models.cloud_collaborators_query import CloudCollaboratorsQuery
from openapi_client.models.cloud_config import CloudConfig
from openapi_client.models.cloud_data_bucket_file_type import CloudDataBucketFileType
from openapi_client.models.cloud_data_bucket_presigned_upload_info import CloudDataBucketPresignedUploadInfo
from openapi_client.models.cloud_data_bucket_presigned_upload_request import CloudDataBucketPresignedUploadRequest
from openapi_client.models.cloud_list_response import CloudListResponse
from openapi_client.models.cloud_name_options import CloudNameOptions
from openapi_client.models.cloud_project_collaborator import CloudProjectCollaborator
from openapi_client.models.cloud_project_collaborator_value import CloudProjectCollaboratorValue
from openapi_client.models.cloud_provider import CloudProvider
from openapi_client.models.cloud_providers import CloudProviders
from openapi_client.models.cloud_region_and_zones import CloudRegionAndZones
from openapi_client.models.cloud_region_info import CloudRegionInfo
from openapi_client.models.cloud_resource import CloudResource
from openapi_client.models.cloud_resource_gcp import CloudResourceGCP
from openapi_client.models.cloud_response import CloudResponse
from openapi_client.models.cloud_state import CloudState
from openapi_client.models.cloud_status import CloudStatus
from openapi_client.models.cloud_types import CloudTypes
from openapi_client.models.cloud_version import CloudVersion
from openapi_client.models.cloud_waitlist_status import CloudWaitlistStatus
from openapi_client.models.cloud_with_cloud_resource import CloudWithCloudResource
from openapi_client.models.cloud_with_cloud_resource_gcp import CloudWithCloudResourceGCP
from openapi_client.models.cloudcollaborator_list_response import CloudcollaboratorListResponse
from openapi_client.models.clouddatabucketpresigneduploadinfo_response import ClouddatabucketpresigneduploadinfoResponse
from openapi_client.models.cloudregionandzones_response import CloudregionandzonesResponse
from openapi_client.models.cloudresource_response import CloudresourceResponse
from openapi_client.models.cloudresourcegcp_response import CloudresourcegcpResponse
from openapi_client.models.cloudwithcloudresource_response import CloudwithcloudresourceResponse
from openapi_client.models.cloudwithcloudresourcegcp_response import CloudwithcloudresourcegcpResponse
from openapi_client.models.cluster_auth_response import ClusterAuthResponse
from openapi_client.models.cluster_config import ClusterConfig
from openapi_client.models.cluster_config_with_session_idle_timeout import ClusterConfigWithSessionIdleTimeout
from openapi_client.models.cluster_environments_query import ClusterEnvironmentsQuery
from openapi_client.models.cluster_event import ClusterEvent
from openapi_client.models.cluster_events_output import ClusterEventsOutput
from openapi_client.models.cluster_features import ClusterFeatures
from openapi_client.models.cluster_management_stack_versions import ClusterManagementStackVersions
from openapi_client.models.cluster_startup import ClusterStartup
from openapi_client.models.cluster_status import ClusterStatus
from openapi_client.models.cluster_status_details import ClusterStatusDetails
from openapi_client.models.clusterauthresponse_response import ClusterauthresponseResponse
from openapi_client.models.clusterconfig_response import ClusterconfigResponse
from openapi_client.models.clusterconfigwithsessionidletimeout_response import ClusterconfigwithsessionidletimeoutResponse
from openapi_client.models.clustereventsoutput_response import ClustereventsoutputResponse
from openapi_client.models.clusterfeatures_response import ClusterfeaturesResponse
from openapi_client.models.company_size import CompanySize
from openapi_client.models.compute_node_type import ComputeNodeType
from openapi_client.models.compute_template import ComputeTemplate
from openapi_client.models.compute_template_config import ComputeTemplateConfig
from openapi_client.models.compute_template_query import ComputeTemplateQuery
from openapi_client.models.computetemplate_response import ComputetemplateResponse
from openapi_client.models.computetemplateconfig_response import ComputetemplateconfigResponse
from openapi_client.models.create_aica_endpoint import CreateAicaEndpoint
from openapi_client.models.create_aioa_cloud_waitlist import CreateAioaCloudWaitlist
from openapi_client.models.create_analytics_event import CreateAnalyticsEvent
from openapi_client.models.create_app_config import CreateAppConfig
from openapi_client.models.create_app_config_configuration_schema import CreateAppConfigConfigurationSchema
from openapi_client.models.create_byod_app_config import CreateBYODAppConfig
from openapi_client.models.create_byod_app_config_configuration_schema import CreateBYODAppConfigConfigurationSchema
from openapi_client.models.create_byod_build import CreateBYODBuild
from openapi_client.models.create_billing_version import CreateBillingVersion
from openapi_client.models.create_bug_report_response import CreateBugReportResponse
from openapi_client.models.create_build import CreateBuild
from openapi_client.models.create_cloud_collaborator import CreateCloudCollaborator
from openapi_client.models.create_cloud_resource import CreateCloudResource
from openapi_client.models.create_cloud_resource_gcp import CreateCloudResourceGCP
from openapi_client.models.create_cloud_with_cloud_resource import CreateCloudWithCloudResource
from openapi_client.models.create_cluster_compute_config import CreateClusterComputeConfig
from openapi_client.models.create_compute_template import CreateComputeTemplate
from openapi_client.models.create_compute_template_config import CreateComputeTemplateConfig
from openapi_client.models.create_experimental_workspace import CreateExperimentalWorkspace
from openapi_client.models.create_experimental_workspace_from_job import CreateExperimentalWorkspaceFromJob
from openapi_client.models.create_fine_tuning_hyperparameters import CreateFineTuningHyperparameters
from openapi_client.models.create_fine_tuning_job_product_request import CreateFineTuningJobProductRequest
from openapi_client.models.create_internal_production_job import CreateInternalProductionJob
from openapi_client.models.create_job_queue_config import CreateJobQueueConfig
from openapi_client.models.create_machine_pool_request import CreateMachinePoolRequest
from openapi_client.models.create_machine_pool_response import CreateMachinePoolResponse
from openapi_client.models.create_machine_request import CreateMachineRequest
from openapi_client.models.create_machine_response import CreateMachineResponse
from openapi_client.models.create_metronome_webhook_notification import CreateMetronomeWebhookNotification
from openapi_client.models.create_otp_return_api_model import CreateOTPReturnApiModel
from openapi_client.models.create_organization_invitation import CreateOrganizationInvitation
from openapi_client.models.create_production_job_config import CreateProductionJobConfig
from openapi_client.models.create_schedule import CreateSchedule
from openapi_client.models.create_session_from_snapshot_options import CreateSessionFromSnapshotOptions
from openapi_client.models.create_session_in_db import CreateSessionInDb
from openapi_client.models.create_session_response import CreateSessionResponse
from openapi_client.models.create_user import CreateUser
from openapi_client.models.create_user_project_collaborator import CreateUserProjectCollaborator
from openapi_client.models.create_user_project_collaborator_value import CreateUserProjectCollaboratorValue
from openapi_client.models.create_workspace_from_template import CreateWorkspaceFromTemplate
from openapi_client.models.createbugreportresponse_response import CreatebugreportresponseResponse
from openapi_client.models.createcomputetemplateconfig_response import CreatecomputetemplateconfigResponse
from openapi_client.models.createmachinepoolresponse_response import CreatemachinepoolresponseResponse
from openapi_client.models.createmachineresponse_response import CreatemachineresponseResponse
from openapi_client.models.createotpreturnapimodel_response import CreateotpreturnapimodelResponse
from openapi_client.models.createsessionresponse_response import CreatesessionresponseResponse
from openapi_client.models.credit_card_information import CreditCardInformation
from openapi_client.models.customer_alert_status import CustomerAlertStatus
from openapi_client.models.dataplane_services import DataplaneServices
from openapi_client.models.decorated_application_template import DecoratedApplicationTemplate
from openapi_client.models.decorated_build import DecoratedBuild
from openapi_client.models.decorated_compute_template import DecoratedComputeTemplate
from openapi_client.models.decorated_compute_template_config import DecoratedComputeTemplateConfig
from openapi_client.models.decorated_interactive_session import DecoratedInteractiveSession
from openapi_client.models.decorated_job import DecoratedJob
from openapi_client.models.decorated_job_queue import DecoratedJobQueue
from openapi_client.models.decorated_job_submission import DecoratedJobSubmission
from openapi_client.models.decorated_list_service_api_model import DecoratedListServiceAPIModel
from openapi_client.models.decorated_production_job import DecoratedProductionJob
from openapi_client.models.decorated_production_job_state_transition import DecoratedProductionJobStateTransition
from openapi_client.models.decorated_production_service_v2_api_model import DecoratedProductionServiceV2APIModel
from openapi_client.models.decorated_production_service_v2_version_api_model import DecoratedProductionServiceV2VersionAPIModel
from openapi_client.models.decorated_runtime_env import DecoratedRuntimeEnv
from openapi_client.models.decorated_schedule import DecoratedSchedule
from openapi_client.models.decorated_serve_deployment import DecoratedServeDeployment
from openapi_client.models.decorated_service_event_api_model import DecoratedServiceEventAPIModel
from openapi_client.models.decorated_session import DecoratedSession
from openapi_client.models.decorated_support_request import DecoratedSupportRequest
from openapi_client.models.decorated_unified_job import DecoratedUnifiedJob
from openapi_client.models.decoratedapplicationtemplate_list_response import DecoratedapplicationtemplateListResponse
from openapi_client.models.decoratedapplicationtemplate_response import DecoratedapplicationtemplateResponse
from openapi_client.models.decoratedbuild_list_response import DecoratedbuildListResponse
from openapi_client.models.decoratedbuild_response import DecoratedbuildResponse
from openapi_client.models.decoratedcomputetemplate_list_response import DecoratedcomputetemplateListResponse
from openapi_client.models.decoratedcomputetemplate_response import DecoratedcomputetemplateResponse
from openapi_client.models.decoratedinteractivesession_list_response import DecoratedinteractivesessionListResponse
from openapi_client.models.decoratedinteractivesession_response import DecoratedinteractivesessionResponse
from openapi_client.models.decoratedjob_list_response import DecoratedjobListResponse
from openapi_client.models.decoratedjob_response import DecoratedjobResponse
from openapi_client.models.decoratedjobqueue_list_response import DecoratedjobqueueListResponse
from openapi_client.models.decoratedjobqueue_response import DecoratedjobqueueResponse
from openapi_client.models.decoratedjobsubmission_list_response import DecoratedjobsubmissionListResponse
from openapi_client.models.decoratedjobsubmission_response import DecoratedjobsubmissionResponse
from openapi_client.models.decoratedlistserviceapimodel_list_response import DecoratedlistserviceapimodelListResponse
from openapi_client.models.decoratedproductionjob_list_response import DecoratedproductionjobListResponse
from openapi_client.models.decoratedproductionjob_response import DecoratedproductionjobResponse
from openapi_client.models.decoratedproductionjobstatetransition_list_response import DecoratedproductionjobstatetransitionListResponse
from openapi_client.models.decoratedproductionservicev2_apimodel_response import Decoratedproductionservicev2ApimodelResponse
from openapi_client.models.decoratedproductionservicev2_versionapimodel_list_response import Decoratedproductionservicev2VersionapimodelListResponse
from openapi_client.models.decoratedruntimeenv_list_response import DecoratedruntimeenvListResponse
from openapi_client.models.decoratedruntimeenv_response import DecoratedruntimeenvResponse
from openapi_client.models.decoratedschedule_list_response import DecoratedscheduleListResponse
from openapi_client.models.decoratedschedule_response import DecoratedscheduleResponse
from openapi_client.models.decoratedservedeployment_list_response import DecoratedservedeploymentListResponse
from openapi_client.models.decoratedservedeployment_response import DecoratedservedeploymentResponse
from openapi_client.models.decoratedserviceeventapimodel_list_response import DecoratedserviceeventapimodelListResponse
from openapi_client.models.decoratedsession_list_response import DecoratedsessionListResponse
from openapi_client.models.decoratedsession_response import DecoratedsessionResponse
from openapi_client.models.decoratedsupportrequest_list_response import DecoratedsupportrequestListResponse
from openapi_client.models.decoratedsupportrequest_response import DecoratedsupportrequestResponse
from openapi_client.models.decoratedunifiedjob_list_response import DecoratedunifiedjobListResponse
from openapi_client.models.decoratedunifiedjob_response import DecoratedunifiedjobResponse
from openapi_client.models.delete_machine_pool_request import DeleteMachinePoolRequest
from openapi_client.models.delete_machine_request import DeleteMachineRequest
from openapi_client.models.deletemachinepoolresponse_response import DeletemachinepoolresponseResponse
from openapi_client.models.detach_machine_pool_from_cloud_request import DetachMachinePoolFromCloudRequest
from openapi_client.models.detachmachinepoolfromcloudresponse_response import DetachmachinepoolfromcloudresponseResponse
from openapi_client.models.dismissal_type import DismissalType
from openapi_client.models.ebs_block_device import EbsBlockDevice
from openapi_client.models.editable_cloud_resource import EditableCloudResource
from openapi_client.models.editable_cloud_resource_gcp import EditableCloudResourceGCP
from openapi_client.models.error import Error
from openapi_client.models.event_level import EventLevel
from openapi_client.models.execute_command_response import ExecuteCommandResponse
from openapi_client.models.execute_interactive_command_options import ExecuteInteractiveCommandOptions
from openapi_client.models.execute_shell_command_options import ExecuteShellCommandOptions
from openapi_client.models.executecommandresponse_response import ExecutecommandresponseResponse
from openapi_client.models.experimental_workspace import ExperimentalWorkspace
from openapi_client.models.experimentalworkspace_list_response import ExperimentalworkspaceListResponse
from openapi_client.models.experimentalworkspace_response import ExperimentalworkspaceResponse
from openapi_client.models.external_service_status import ExternalServiceStatus
from openapi_client.models.external_service_status_response import ExternalServiceStatusResponse
from openapi_client.models.external_terminal_command import ExternalTerminalCommand
from openapi_client.models.externalservicestatusresponse_response import ExternalservicestatusresponseResponse
from openapi_client.models.feature_compatibility import FeatureCompatibility
from openapi_client.models.feature_flag_response import FeatureFlagResponse
from openapi_client.models.featureflagresponse_response import FeatureflagresponseResponse
from openapi_client.models.file_model import FileModel
from openapi_client.models.filemodel_response import FilemodelResponse
from openapi_client.models.fine_tuning_job_status import FineTuningJobStatus
from openapi_client.models.finish_ft_job_request import FinishFTJobRequest
from openapi_client.models.finish_ft_job_request_v2 import FinishFTJobRequestV2
from openapi_client.models.gcp_file_store_config import GCPFileStoreConfig
from openapi_client.models.gcp_memorystore_instance_config import GCPMemorystoreInstanceConfig
from openapi_client.models.gcp_node_disk import GCPNodeDisk
from openapi_client.models.gcp_node_options import GCPNodeOptions
from openapi_client.models.grafana_dashboard import GrafanaDashboard
from openapi_client.models.grpc_protocol_config import GrpcProtocolConfig
from openapi_client.models.http_validation_error import HTTPValidationError
from openapi_client.models.ha_job_error_types import HaJobErrorTypes
from openapi_client.models.ha_job_event_level import HaJobEventLevel
from openapi_client.models.ha_job_event_origin import HaJobEventOrigin
from openapi_client.models.ha_job_goal_states import HaJobGoalStates
from openapi_client.models.ha_job_states import HaJobStates
from openapi_client.models.ha_job_type import HaJobType
from openapi_client.models.ha_jobs_sort_field import HaJobsSortField
from openapi_client.models.head_ip import HeadIp
from openapi_client.models.headip_response import HeadipResponse
from openapi_client.models.http_protocol_config import HttpProtocolConfig
from openapi_client.models.iam_instance_profile_specification import IamInstanceProfileSpecification
from openapi_client.models.idle_termination_status import IdleTerminationStatus
from openapi_client.models.import_aica_model import ImportAicaModel
from openapi_client.models.integration_details import IntegrationDetails
from openapi_client.models.interactive_session_logs import InteractiveSessionLogs
from openapi_client.models.interactivesessionlogs_response import InteractivesessionlogsResponse
from openapi_client.models.internal_production_job import InternalProductionJob
from openapi_client.models.internalproductionjob_response import InternalproductionjobResponse
from openapi_client.models.invoice import Invoice
from openapi_client.models.invoice_list_response import InvoiceListResponse
from openapi_client.models.invoice_status import InvoiceStatus
from openapi_client.models.invoices_query import InvoicesQuery
from openapi_client.models.job_access import JobAccess
from openapi_client.models.job_queue_config import JobQueueConfig
from openapi_client.models.job_queue_execution_mode import JobQueueExecutionMode
from openapi_client.models.job_queue_spec import JobQueueSpec
from openapi_client.models.job_queue_state import JobQueueState
from openapi_client.models.job_queues_query import JobQueuesQuery
from openapi_client.models.job_run_type import JobRunType
from openapi_client.models.job_state_log_level_types import JobStateLogLevelTypes
from openapi_client.models.job_status import JobStatus
from openapi_client.models.job_submissions_sort_field import JobSubmissionsSortField
from openapi_client.models.jobs_logs import JobsLogs
from openapi_client.models.jobs_logs_query_info import JobsLogsQueryInfo
from openapi_client.models.jobs_sort_field import JobsSortField
from openapi_client.models.jobslogs_response import JobslogsResponse
from openapi_client.models.jobslogsqueryinfo_response import JobslogsqueryinfoResponse
from openapi_client.models.json_patch_operation import JsonPatchOperation
from openapi_client.models.lb_resource import LBResource
from openapi_client.models.lbresource_response import LbresourceResponse
from openapi_client.models.list_machine_pools_response import ListMachinePoolsResponse
from openapi_client.models.list_machines_response import ListMachinesResponse
from openapi_client.models.list_response_metadata import ListResponseMetadata
from openapi_client.models.listmachinepoolsresponse_response import ListmachinepoolsresponseResponse
from openapi_client.models.listmachinesresponse_response import ListmachinesresponseResponse
from openapi_client.models.log_detail import LogDetail
from openapi_client.models.log_details import LogDetails
from openapi_client.models.log_download_config import LogDownloadConfig
from openapi_client.models.log_download_request import LogDownloadRequest
from openapi_client.models.log_download_result import LogDownloadResult
from openapi_client.models.log_file_chunk import LogFileChunk
from openapi_client.models.log_filter import LogFilter
from openapi_client.models.log_item import LogItem
from openapi_client.models.log_item_batch import LogItemBatch
from openapi_client.models.log_level_types import LogLevelTypes
from openapi_client.models.log_stream import LogStream
from openapi_client.models.logdetails_response import LogdetailsResponse
from openapi_client.models.logdownloadresult_response import LogdownloadresultResponse
from openapi_client.models.login_user_params import LoginUserParams
from openapi_client.models.logitembatch_response import LogitembatchResponse
from openapi_client.models.logs_output import LogsOutput
from openapi_client.models.logsoutput_response import LogsoutputResponse
from openapi_client.models.logstream_response import LogstreamResponse
from openapi_client.models.machine_allocation_state import MachineAllocationState
from openapi_client.models.machine_connection_state import MachineConnectionState
from openapi_client.models.machine_info import MachineInfo
from openapi_client.models.machine_pool import MachinePool
from openapi_client.models.metronome_dashboard_type import MetronomeDashboardType
from openapi_client.models.mini_build import MiniBuild
from openapi_client.models.mini_cloud import MiniCloud
from openapi_client.models.mini_cluster import MiniCluster
from openapi_client.models.mini_compute_template import MiniComputeTemplate
from openapi_client.models.mini_compute_template_config import MiniComputeTemplateConfig
from openapi_client.models.mini_job_run import MiniJobRun
from openapi_client.models.mini_namespace import MiniNamespace
from openapi_client.models.mini_organization import MiniOrganization
from openapi_client.models.mini_production_job import MiniProductionJob
from openapi_client.models.mini_project import MiniProject
from openapi_client.models.mini_runtime_environment import MiniRuntimeEnvironment
from openapi_client.models.mini_schedule import MiniSchedule
from openapi_client.models.mini_user import MiniUser
from openapi_client.models.minibuild_list_response import MinibuildListResponse
from openapi_client.models.minicomputetemplate_list_response import MinicomputetemplateListResponse
from openapi_client.models.miniproject_list_response import MiniprojectListResponse
from openapi_client.models.monitor_logs_extension import MonitorLogsExtension
from openapi_client.models.named_entity import NamedEntity
from openapi_client.models.network_interface import NetworkInterface
from openapi_client.models.node_registration_aws import NodeRegistrationAWS
from openapi_client.models.node_registration_gcp import NodeRegistrationGCP
from openapi_client.models.node_registration_provisioned import NodeRegistrationProvisioned
from openapi_client.models.node_registration_sagemaker import NodeRegistrationSagemaker
from openapi_client.models.node_registration_v2 import NodeRegistrationV2
from openapi_client.models.node_type import NodeType
from openapi_client.models.onboarding_user_cards_query import OnboardingUserCardsQuery
from openapi_client.models.organization import Organization
from openapi_client.models.organization_availability import OrganizationAvailability
from openapi_client.models.organization_collaborator import OrganizationCollaborator
from openapi_client.models.organization_invitation import OrganizationInvitation
from openapi_client.models.organization_invitation_base import OrganizationInvitationBase
from openapi_client.models.organization_marketing_questions import OrganizationMarketingQuestions
from openapi_client.models.organization_permission_level import OrganizationPermissionLevel
from openapi_client.models.organization_project_collaborator import OrganizationProjectCollaborator
from openapi_client.models.organization_project_collaborator_value import OrganizationProjectCollaboratorValue
from openapi_client.models.organization_public_identifier import OrganizationPublicIdentifier
from openapi_client.models.organization_response import OrganizationResponse
from openapi_client.models.organization_summary import OrganizationSummary
from openapi_client.models.organization_usage_alert import OrganizationUsageAlert
from openapi_client.models.organization_usage_alert_severity import OrganizationUsageAlertSeverity
from openapi_client.models.organizationavailability_response import OrganizationavailabilityResponse
from openapi_client.models.organizationcollaborator_list_response import OrganizationcollaboratorListResponse
from openapi_client.models.organizationinvitation_list_response import OrganizationinvitationListResponse
from openapi_client.models.organizationinvitation_response import OrganizationinvitationResponse
from openapi_client.models.organizationinvitationbase_response import OrganizationinvitationbaseResponse
from openapi_client.models.organizationprojectcollaborator_list_response import OrganizationprojectcollaboratorListResponse
from openapi_client.models.organizationpublicidentifier_response import OrganizationpublicidentifierResponse
from openapi_client.models.organizationusagealert_list_response import OrganizationusagealertListResponse
from openapi_client.models.page_query import PageQuery
from openapi_client.models.pause_schedule import PauseSchedule
from openapi_client.models.permission_level import PermissionLevel
from openapi_client.models.platform_fine_tuning_job import PlatformFineTuningJob
from openapi_client.models.platformfinetuningjob_list_response import PlatformfinetuningjobListResponse
from openapi_client.models.platformfinetuningjob_response import PlatformfinetuningjobResponse
from openapi_client.models.presigned_upload_file import PresignedUploadFile
from openapi_client.models.presigned_upload_file_request import PresignedUploadFileRequest
from openapi_client.models.product_autoscaler_flag import ProductAutoscalerFlag
from openapi_client.models.product_type import ProductType
from openapi_client.models.productautoscalerflag_response import ProductautoscalerflagResponse
from openapi_client.models.production_job import ProductionJob
from openapi_client.models.production_job_config import ProductionJobConfig
from openapi_client.models.production_job_event import ProductionJobEvent
from openapi_client.models.production_job_event_scope_filter import ProductionJobEventScopeFilter
from openapi_client.models.production_job_state_transition import ProductionJobStateTransition
from openapi_client.models.productionjob_response import ProductionjobResponse
from openapi_client.models.productionjobevent_list_response import ProductionjobeventListResponse
from openapi_client.models.project import Project
from openapi_client.models.project_base import ProjectBase
from openapi_client.models.project_collaborator import ProjectCollaborator
from openapi_client.models.project_collaborator_value import ProjectCollaboratorValue
from openapi_client.models.project_collaborators_put_message import ProjectCollaboratorsPutMessage
from openapi_client.models.project_create_message import ProjectCreateMessage
from openapi_client.models.project_default_session_name import ProjectDefaultSessionName
from openapi_client.models.project_delete_message import ProjectDeleteMessage
from openapi_client.models.project_list_response import ProjectListResponse
from openapi_client.models.project_patch_message import ProjectPatchMessage
from openapi_client.models.project_response import ProjectResponse
from openapi_client.models.projectbase_response import ProjectbaseResponse
from openapi_client.models.projectcollaborator_list_response import ProjectcollaboratorListResponse
from openapi_client.models.projectdefaultsessionname_response import ProjectdefaultsessionnameResponse
from openapi_client.models.projects_sort_field import ProjectsSortField
from openapi_client.models.projects_violating_tree_hierarchy_response import ProjectsViolatingTreeHierarchyResponse
from openapi_client.models.projectsviolatingtreehierarchyresponse_response import ProjectsviolatingtreehierarchyresponseResponse
from openapi_client.models.protocols import Protocols
from openapi_client.models.provider_metadata import ProviderMetadata
from openapi_client.models.providermetadata_response import ProvidermetadataResponse
from openapi_client.models.python_modules import PythonModules
from openapi_client.models.ray_gcs_external_storage_config import RayGCSExternalStorageConfig
from openapi_client.models.ray_runtime_env_config import RayRuntimeEnvConfig
from openapi_client.models.read_billing_version import ReadBillingVersion
from openapi_client.models.readbillingversion_list_response import ReadbillingversionListResponse
from openapi_client.models.replica_details import ReplicaDetails
from openapi_client.models.replica_state import ReplicaState
from openapi_client.models.request_email_magic_link_response import RequestEmailMagicLinkResponse
from openapi_client.models.request_otp_return_api_model import RequestOTPReturnApiModel
from openapi_client.models.request_password_reset_params import RequestPasswordResetParams
from openapi_client.models.requestemailmagiclinkresponse_response import RequestemailmagiclinkresponseResponse
from openapi_client.models.requestotpreturnapimodel_response import RequestotpreturnapimodelResponse
from openapi_client.models.reset_password_params import ResetPasswordParams
from openapi_client.models.resources import Resources
from openapi_client.models.resubmit_ft_job_request import ResubmitFTJobRequest
from openapi_client.models.rollback_service_model import RollbackServiceModel
from openapi_client.models.rollout_strategy import RolloutStrategy
from openapi_client.models.s3_download_location import S3DownloadLocation
from openapi_client.models.sso_login_info import SSOLoginInfo
from openapi_client.models.supportedbaseimagesenum import SUPPORTEDBASEIMAGESENUM
from openapi_client.models.schedule_config import ScheduleConfig
from openapi_client.models.serve_deployment_grafana_dashboard_status import ServeDeploymentGrafanaDashboardStatus
from openapi_client.models.serve_deployment_logs import ServeDeploymentLogs
from openapi_client.models.serve_deployment_state import ServeDeploymentState
from openapi_client.models.servedeploymentlogs_response import ServedeploymentlogsResponse
from openapi_client.models.server_session_token import ServerSessionToken
from openapi_client.models.serversessiontoken_response import ServersessiontokenResponse
from openapi_client.models.service_account import ServiceAccount
from openapi_client.models.service_config import ServiceConfig
from openapi_client.models.service_event_current_state import ServiceEventCurrentState
from openapi_client.models.service_event_level import ServiceEventLevel
from openapi_client.models.service_event_origin import ServiceEventOrigin
from openapi_client.models.service_event_scope import ServiceEventScope
from openapi_client.models.service_event_scope_filter import ServiceEventScopeFilter
from openapi_client.models.service_event_type import ServiceEventType
from openapi_client.models.service_event_verbose_message_model import ServiceEventVerboseMessageModel
from openapi_client.models.service_goal_states import ServiceGoalStates
from openapi_client.models.service_observability_urls import ServiceObservabilityUrls
from openapi_client.models.service_sort_field import ServiceSortField
from openapi_client.models.service_type import ServiceType
from openapi_client.models.service_usage import ServiceUsage
from openapi_client.models.service_version_state import ServiceVersionState
from openapi_client.models.serviceeventverbosemessagemodel_list_response import ServiceeventverbosemessagemodelListResponse
from openapi_client.models.session import Session
from openapi_client.models.session_access import SessionAccess
from openapi_client.models.session_autosync_sessions_update_message import SessionAutosyncSessionsUpdateMessage
from openapi_client.models.session_command import SessionCommand
from openapi_client.models.session_command_finish_options import SessionCommandFinishOptions
from openapi_client.models.session_command_id import SessionCommandId
from openapi_client.models.session_command_types import SessionCommandTypes
from openapi_client.models.session_create_message import SessionCreateMessage
from openapi_client.models.session_delete_message import SessionDeleteMessage
from openapi_client.models.session_describe import SessionDescribe
from openapi_client.models.session_details import SessionDetails
from openapi_client.models.session_event import SessionEvent
from openapi_client.models.session_event_cause import SessionEventCause
from openapi_client.models.session_event_types import SessionEventTypes
from openapi_client.models.session_execute_message import SessionExecuteMessage
from openapi_client.models.session_finish_command_message import SessionFinishCommandMessage
from openapi_client.models.session_history_item import SessionHistoryItem
from openapi_client.models.session_kill_command_message import SessionKillCommandMessage
from openapi_client.models.session_list_response import SessionListResponse
from openapi_client.models.session_patch_message import SessionPatchMessage
from openapi_client.models.session_response import SessionResponse
from openapi_client.models.session_ssh_key import SessionSshKey
from openapi_client.models.session_starting_up_data import SessionStartingUpData
from openapi_client.models.session_state import SessionState
from openapi_client.models.session_state_change_message import SessionStateChangeMessage
from openapi_client.models.session_state_data import SessionStateData
from openapi_client.models.session_stopping_data import SessionStoppingData
from openapi_client.models.sessioncommand_list_response import SessioncommandListResponse
from openapi_client.models.sessioncommandid_response import SessioncommandidResponse
from openapi_client.models.sessiondescribe_response import SessiondescribeResponse
from openapi_client.models.sessiondetails_response import SessiondetailsResponse
from openapi_client.models.sessionevent_list_response import SessioneventListResponse
from openapi_client.models.sessionhistoryitem_list_response import SessionhistoryitemListResponse
from openapi_client.models.sessions_sort_field import SessionsSortField
from openapi_client.models.sessionsshkey_response import SessionsshkeyResponse
from openapi_client.models.setup_initialize_session_options import SetupInitializeSessionOptions
from openapi_client.models.show_otp_source_return_api_model import ShowOTPSourceReturnApiModel
from openapi_client.models.showotpsourcereturnapimodel_response import ShowotpsourcereturnapimodelResponse
from openapi_client.models.snapshot_create_message import SnapshotCreateMessage
from openapi_client.models.snapshot_delete_message import SnapshotDeleteMessage
from openapi_client.models.snapshot_patch_message import SnapshotPatchMessage
from openapi_client.models.socket_message_schemas import SocketMessageSchemas
from openapi_client.models.socket_message_types import SocketMessageTypes
from openapi_client.models.socketmessageschemas_response import SocketmessageschemasResponse
from openapi_client.models.socketmessagetypes_response import SocketmessagetypesResponse
from openapi_client.models.sort_order import SortOrder
from openapi_client.models.ssologininfo_response import SsologininfoResponse
from openapi_client.models.start_session_options import StartSessionOptions
from openapi_client.models.stop_session_options import StopSessionOptions
from openapi_client.models.stream_publish_request import StreamPublishRequest
from openapi_client.models.subnet_id_with_availability_zone_aws import SubnetIdWithAvailabilityZoneAWS
from openapi_client.models.support_requests_query import SupportRequestsQuery
from openapi_client.models.templatized_compute_configs import TemplatizedComputeConfigs
from openapi_client.models.templatized_decorated_application_templates import TemplatizedDecoratedApplicationTemplates
from openapi_client.models.templatizedcomputeconfigs_response import TemplatizedcomputeconfigsResponse
from openapi_client.models.templatizeddecoratedapplicationtemplates_response import TemplatizeddecoratedapplicationtemplatesResponse
from openapi_client.models.text_query import TextQuery
from openapi_client.models.timestamped_logs_output import TimestampedLogsOutput
from openapi_client.models.timestampedlogsoutput_response import TimestampedlogsoutputResponse
from openapi_client.models.tool import Tool
from openapi_client.models.tracing_config import TracingConfig
from openapi_client.models.try_login_email_response import TryLoginEmailResponse
from openapi_client.models.tryloginemailresponse_response import TryloginemailresponseResponse
from openapi_client.models.unified_job_sort_field import UnifiedJobSortField
from openapi_client.models.unified_job_status import UnifiedJobStatus
from openapi_client.models.unified_job_type import UnifiedJobType
from openapi_client.models.update_cloud_with_cloud_resource import UpdateCloudWithCloudResource
from openapi_client.models.update_cloud_with_cloud_resource_gcp import UpdateCloudWithCloudResourceGCP
from openapi_client.models.update_cluster_dns import UpdateClusterDns
from openapi_client.models.update_compute_template import UpdateComputeTemplate
from openapi_client.models.update_compute_template_config import UpdateComputeTemplateConfig
from openapi_client.models.update_endpoint import UpdateEndpoint
from openapi_client.models.update_model_deployment import UpdateModelDeployment
from openapi_client.models.update_organization_collaborator import UpdateOrganizationCollaborator
from openapi_client.models.update_project_collaborator import UpdateProjectCollaborator
from openapi_client.models.upload_session_command_logs_locations import UploadSessionCommandLogsLocations
from openapi_client.models.uploadsessioncommandlogslocations_response import UploadsessioncommandlogslocationsResponse
from openapi_client.models.user_info import UserInfo
from openapi_client.models.user_resend_email_options import UserResendEmailOptions
from openapi_client.models.user_service_access_types import UserServiceAccessTypes
from openapi_client.models.userinfo_response import UserinfoResponse
from openapi_client.models.utm_fields import UtmFields
from openapi_client.models.validate_otp_params_api_model import ValidateOTPParamsApiModel
from openapi_client.models.validation_error import ValidationError
from openapi_client.models.verify_response import VerifyResponse
from openapi_client.models.verifyresponse_response import VerifyresponseResponse
from openapi_client.models.visibility import Visibility
from openapi_client.models.waitlist_status_response import WaitlistStatusResponse
from openapi_client.models.waitlist_status_type import WaitlistStatusType
from openapi_client.models.waitliststatusresponse_response import WaitliststatusresponseResponse
from openapi_client.models.wand_b_run_details import WandBRunDetails
from openapi_client.models.web_terminal import WebTerminal
from openapi_client.models.webterminal_list_response import WebterminalListResponse
from openapi_client.models.webterminal_response import WebterminalResponse
from openapi_client.models.worker_node_type import WorkerNodeType
from openapi_client.models.workspace_dataplane_artifact import WorkspaceDataplaneArtifact
from openapi_client.models.workspace_dataplane_artifacts import WorkspaceDataplaneArtifacts
from openapi_client.models.workspace_dataplane_proxied_artifacts import WorkspaceDataplaneProxiedArtifacts
from openapi_client.models.workspace_event import WorkspaceEvent
from openapi_client.models.workspace_event_source import WorkspaceEventSource
from openapi_client.models.workspace_event_source_filter import WorkspaceEventSourceFilter
from openapi_client.models.workspace_readme import WorkspaceReadme
from openapi_client.models.workspace_snapshot_states import WorkspaceSnapshotStates
from openapi_client.models.workspace_template import WorkspaceTemplate
from openapi_client.models.workspace_template_cluster_environment_metadata import WorkspaceTemplateClusterEnvironmentMetadata
from openapi_client.models.workspacedataplaneartifacts_response import WorkspacedataplaneartifactsResponse
from openapi_client.models.workspacedataplaneproxiedartifacts_response import WorkspacedataplaneproxiedartifactsResponse
from openapi_client.models.workspaceevent_list_response import WorkspaceeventListResponse
from openapi_client.models.workspacereadme_response import WorkspacereadmeResponse
from openapi_client.models.workspacetemplate_list_response import WorkspacetemplateListResponse
from openapi_client.models.workspacetemplateclusterenvironmentmetadata_response import WorkspacetemplateclusterenvironmentmetadataResponse
from openapi_client.models.write_cloud import WriteCloud
from openapi_client.models.write_cluster_config import WriteClusterConfig
from openapi_client.models.write_project import WriteProject
from openapi_client.models.write_session import WriteSession
from openapi_client.models.write_support_request import WriteSupportRequest



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import itertools
from typing import Callable, List, Optional, Tuple, TypeVar


T = TypeVar("T")


def partition(
    items: List[T], predicate: Callable[[T], bool]
) -> Tuple[List[T], List[T]]:
    """Partitions provided list into 2 based on the value of the predicate, where:
        - First element is the list of elements for which value of the predicate is truthful
        - Second element are remaining elements
    """
    first = []
    second = []
    for i in items:
        if predicate(i):
            first.append(i)
        else:
            second.append(i)

    return first, second


def flatten(*lists: List[T]) -> List[T]:
    return list(itertools.chain(*lists))


def as_list(o: Optional[T]) -> List[T]:
    """Converts "container" holding optional value into a list"""
    return [o] if o else []



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CreateBYODAppConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'config_json': 'CreateBYODAppConfigConfigurationSchema',
        'anonymous': 'bool'
    }

    attribute_map = {
        'name': 'name',
        'config_json': 'config_json',
        'anonymous': 'anonymous'
    }

    def __init__(self, name=None, config_json=None, anonymous=False, local_vars_configuration=None):  # noqa: E501
        """CreateBYODAppConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._config_json = None
        self._anonymous = None
        self.discriminator = None

        self.name = name
        self.config_json = config_json
        if anonymous is not None:
            self.anonymous = anonymous

    @property
    def name(self):
        """Gets the name of this CreateBYODAppConfig.  # noqa: E501

        Name of the App Template.  # noqa: E501

        :return: The name of this CreateBYODAppConfig.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this CreateBYODAppConfig.

        Name of the App Template.  # noqa: E501

        :param name: The name of this CreateBYODAppConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def config_json(self):
        """Gets the config_json of this CreateBYODAppConfig.  # noqa: E501

        Config JSON to use to create a new BYOD App Config.  # noqa: E501

        :return: The config_json of this CreateBYODAppConfig.  # noqa: E501
        :rtype: CreateBYODAppConfigConfigurationSchema
        """
        return self._config_json

    @config_json.setter
    def config_json(self, config_json):
        """Sets the config_json of this CreateBYODAppConfig.

        Config JSON to use to create a new BYOD App Config.  # noqa: E501

        :param config_json: The config_json of this CreateBYODAppConfig.  # noqa: E501
        :type: CreateBYODAppConfigConfigurationSchema
        """
        if self.local_vars_configuration.client_side_validation and config_json is None:  # noqa: E501
            raise ValueError("Invalid value for `config_json`, must not be `None`")  # noqa: E501

        self._config_json = config_json

    @property
    def anonymous(self):
        """Gets the anonymous of this CreateBYODAppConfig.  # noqa: E501

        True if this is an anonymous app config.  # noqa: E501

        :return: The anonymous of this CreateBYODAppConfig.  # noqa: E501
        :rtype: bool
        """
        return self._anonymous

    @anonymous.setter
    def anonymous(self, anonymous):
        """Sets the anonymous of this CreateBYODAppConfig.

        True if this is an anonymous app config.  # noqa: E501

        :param anonymous: The anonymous of this CreateBYODAppConfig.  # noqa: E501
        :type: bool
        """

        self._anonymous = anonymous

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateBYODAppConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateBYODAppConfig):
            return True

        return self.to_dict() != other.to_dict()



import os
import tempfile
from unittest.mock import Mock, patch

import pytest

from anyscale.client.openapi_client.models import (
    SessionSshKey,
    SessionsshkeyResponse,
)
from anyscale.controllers.workspace_controller import WorkspaceController
from anyscale.feature_flags import FLAG_DEFAULT_WORKING_DIR_FOR_PROJ
from anyscale.utils.workspace_utils import _get_ssh_key_info


def test_configure_ssh_key(base_mock_api_client: Mock) -> None:
    base_mock_api_client.get_session_ssh_key_api_v2_sessions_session_id_ssh_key_get = Mock(
        return_value=SessionsshkeyResponse(
            result=SessionSshKey(key_name="SSH_KEY", private_key="PRIVATE_KEY")
        )
    )
    with tempfile.TemporaryDirectory() as directory:
        _get_ssh_key_info("session_id", base_mock_api_client, directory)
        with open(os.path.join(directory, "SSH_KEY.pem")) as f:
            assert f.read() == "PRIVATE_KEY"


@pytest.mark.parametrize("default_working_dir_for_proj_flag", [True, False])
def test_get_workspace_dir_name(
    mock_auth_api_client, default_working_dir_for_proj_flag: bool
):
    workspace_controller = WorkspaceController()
    workspace_controller.api_client = Mock()

    mock_workspace = Mock(project_id="mock_project_id")
    mock_project = Mock(directory_name="default")
    mock_project.name = "default_cld_1"
    workspace_controller.api_client.get_workspace_api_v2_experimental_workspaces_workspace_id_get = Mock(
        return_value=Mock(result=mock_workspace)
    )
    workspace_controller.api_client.get_project_api_v2_projects_project_id_get = Mock(
        return_value=Mock(result=mock_project)
    )
    workspace_controller.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get = Mock(
        return_value=Mock(result=Mock(is_on=default_working_dir_for_proj_flag))
    )

    with patch.multiple(
        "anyscale.controllers.workspace_controller",
        load_workspace_id_or_throw=Mock(return_value="mock_workspace_id"),
    ):
        directory_name = workspace_controller.get_workspace_dir_name()

    assert (
        directory_name == "default"
        if default_working_dir_for_proj_flag
        else "default_cld_1"
    )

    workspace_controller.api_client.get_workspace_api_v2_experimental_workspaces_workspace_id_get.assert_called_once_with(
        "mock_workspace_id"
    )
    workspace_controller.api_client.get_project_api_v2_projects_project_id_get.assert_called_once_with(
        "mock_project_id"
    )
    workspace_controller.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get.assert_called_once_with(
        FLAG_DEFAULT_WORKING_DIR_FOR_PROJ
    )



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

from anyscale._private.models import ImageURI, ModelBase
from anyscale.compute_config.models import ComputeConfig


@dataclass(frozen=True)
class WorkloadConfig(ModelBase):
    name: Optional[str] = field(
        default=None, metadata={"docstring": "Should be overwritten by subclass."}
    )

    def _validate_name(self, name: Optional[str]):
        if name is not None and not isinstance(name, str):
            raise TypeError("'name' must be a string.")

    image_uri: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "URI of an existing image. Exclusive with `containerfile`."
        },
    )

    def _validate_image_uri(self, image_uri: Optional[str]):
        if image_uri is not None and not isinstance(image_uri, str):
            raise TypeError(f"'image_uri' must be an str but it is {type(image_uri)}.")
        if image_uri is not None:
            ImageURI.from_str(image_uri)

    containerfile: Optional[str] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "The file path to a containerfile that will be built into an image before running the workload. Exclusive with `image_uri`."
        },
    )

    def _validate_containerfile(self, containerfile: Optional[str]):
        if containerfile is not None and not isinstance(containerfile, str):
            raise TypeError("'containerfile' must be a string.")

    compute_config: Union[ComputeConfig, Dict, str, None] = field(
        default=None,
        metadata={
            "docstring": "The name of an existing registered compute config or an inlined ComputeConfig object."
        },
    )

    def _validate_compute_config(
        self, compute_config: Union[ComputeConfig, Dict, str, None]
    ) -> Union[None, str, ComputeConfig]:
        if compute_config is None or isinstance(compute_config, str):
            return compute_config

        if isinstance(compute_config, dict):
            compute_config = ComputeConfig.from_dict(compute_config)
        if not isinstance(compute_config, ComputeConfig):
            raise TypeError(
                "'compute_config' must be a string, ComputeConfig, or corresponding dict"
            )

        return compute_config

    working_dir: Optional[str] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Directory that will be used as the working directory for the application. If a local directory is provided, it will be uploaded to cloud storage automatically. When running inside a workspace, this defaults to the current working directory ('.')."
        },
    )

    def _validate_working_dir(self, working_dir: Optional[str]):
        if working_dir is not None and not isinstance(working_dir, str):
            raise TypeError("'working_dir' must be a string.")

    excludes: Optional[List[str]] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "A list of file path globs that will be excluded when uploading local files for `working_dir`."
        },
    )

    def _validate_excludes(self, excludes: Optional[List[str]]):
        if excludes is not None and (
            not isinstance(excludes, list)
            or not all(isinstance(e, str) for e in excludes)
        ):
            raise TypeError("'excludes' must be a list of strings.")

    requirements: Optional[Union[str, List[str]]] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "A list of pip requirements or a path to a `requirements.txt` file for the workload. When running inside a workspace, this defaults to the workspace-tracked requirements."
        },
    )

    def _validate_requirements(self, requirements: Optional[Union[str, List[str]]]):
        if requirements is None or isinstance(requirements, str):
            return

        if not isinstance(requirements, list) or not all(
            isinstance(r, str) for r in requirements
        ):
            raise TypeError(
                "'requirements' must be a string (file path) or list of strings."
            )

    env_vars: Optional[Dict[str, str]] = field(
        default=None,
        repr=True,
        metadata={
            "docstring": "A dictionary of environment variables that will be set for the workload."
        },
    )

    def _validate_env_vars(self, env_vars: Optional[Dict[str, str]]):
        if env_vars is not None and (
            not isinstance(env_vars, dict)
            or not all(
                isinstance(k, str) and isinstance(v, str) for k, v in env_vars.items()
            )
        ):
            raise TypeError("'env_vars' must be a Dict[str, str].")

    py_modules: Optional[List[str]] = field(
        default=None,
        repr=True,
        metadata={
            "docstring": "A list of local directories or remote URIs that will be uploaded and added to the Python path."
        },
    )

    def _validate_py_modules(self, py_modules: Optional[List[str]]):
        if py_modules is not None and (
            not isinstance(py_modules, list)
            or not all(isinstance(m, str) for m in py_modules)
        ):
            raise TypeError("'py_modules' must be a list of strings.")

    cloud: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace)."
        },
    )

    def _validate_cloud(self, cloud: Optional[str]):
        if cloud is not None and not isinstance(cloud, str):
            raise TypeError("'cloud' must be a string.")

    project: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The project for the workload. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace)."
        },
    )

    def _validate_project(self, project: Optional[str]):
        if project is not None and not isinstance(project, str):
            raise TypeError("'project' must be a string.")

    registry_login_secret: Optional[str] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "A name or identifier of the secret containing credentials to authenticate to the docker registry hosting the image. "
            "This can only be used when 'image_uri' is specified and the image is not hosted on Anyscale."
        },
    )

    def _validate_registry_login_secret(self, registry_login_secret: Optional[str]):
        if registry_login_secret is not None and not isinstance(
            registry_login_secret, str
        ):
            raise TypeError("'registry_login_secret' must be a string.")

    ray_version: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "The Ray version (X.Y.Z) specified for this image specified by either an image URI or a containerfile. If not provided, the latest Ray version will be used."
        },
    )

    def _validate_ray_version(self, ray_version: Optional[str]):
        if ray_version is not None and not isinstance(ray_version, str):
            raise TypeError("'ray_version' must be a string.")
        if ray_version:
            vs = ray_version.split(".")
            if len(vs) != 3:
                raise ValueError(
                    f"Invalid Ray version format: {ray_version}. Must be in the format 'X.Y.Z'."
                )



from typing import Any, Dict, Generator
from unittest.mock import Mock, patch

from click.testing import CliRunner
import pytest

from anyscale.commands.service_commands import rollback, rollout


SERVICE_CONFIG_DICT = {
    "name": "test-service",
    "ray_serve_config": {"applications": [{"import_path": "main:app",},],},
}


class MockServiceController(Mock):
    def read_service_config_file(self, *args, **kwargs) -> Dict[str, Any]:
        return SERVICE_CONFIG_DICT

    def get_service_id(self, service_id: str, **kwargs) -> str:
        return service_id


@pytest.fixture()
def mock_service_controller() -> Generator[MockServiceController, None, None]:
    mock_service_controller = MockServiceController()
    mock_service_controller_cls = Mock(return_value=mock_service_controller,)

    with patch(
        "anyscale.commands.service_commands.ServiceController",
        new=mock_service_controller_cls,
    ):
        yield mock_service_controller


@pytest.fixture()
def mock_service_sdk() -> Generator[MockServiceController, None, None]:
    mock_service_controller = MockServiceController()
    mock_service_controller_cls = Mock(return_value=mock_service_controller,)

    with patch(
        "anyscale.commands.service_commands.ServiceController",
        new=mock_service_controller_cls,
    ):
        yield mock_service_controller


def test_rollout_strategy(mock_service_controller: MockServiceController):
    """Tests the logic for setting `rollout_strategy`.

    This can either be set by `--rollout-strategy` or the `-i / --in-place` alias (but not both).
    """
    runner = CliRunner()
    # --rollout-strategy provided directly.
    runner.invoke(rollout, args=["--rollout-strategy=IN_PLACE", "-f", "file"])
    mock_service_controller.rollout.assert_called_once_with(
        SERVICE_CONFIG_DICT,
        name=None,
        version=None,
        canary_percent=None,
        rollout_strategy="IN_PLACE",
        auto_complete_rollout=True,
        max_surge_percent=None,
    )
    mock_service_controller.reset_mock()

    runner.invoke(rollout, args=["--rollout-strategy=ROLLOUT", "-f", "file"])
    mock_service_controller.rollout.assert_called_once_with(
        SERVICE_CONFIG_DICT,
        name=None,
        version=None,
        canary_percent=None,
        rollout_strategy="ROLLOUT",
        auto_complete_rollout=True,
        max_surge_percent=None,
    )
    mock_service_controller.reset_mock()

    # -i / --in-place used.
    runner.invoke(rollout, args=["-i", "-f", "file"])
    mock_service_controller.rollout.assert_called_once_with(
        SERVICE_CONFIG_DICT,
        name=None,
        version=None,
        canary_percent=None,
        rollout_strategy="IN_PLACE",
        auto_complete_rollout=True,
        max_surge_percent=None,
    )
    mock_service_controller.reset_mock()

    runner.invoke(rollout, args=["--in-place", "-f", "file"])
    mock_service_controller.rollout.assert_called_once_with(
        SERVICE_CONFIG_DICT,
        name=None,
        version=None,
        canary_percent=None,
        rollout_strategy="IN_PLACE",
        auto_complete_rollout=True,
        max_surge_percent=None,
    )
    mock_service_controller.reset_mock()

    runner.invoke(rollout, args=["--max-surge-percent=20", "-f", "file"])
    mock_service_controller.rollout.assert_called_once_with(
        SERVICE_CONFIG_DICT,
        name=None,
        version=None,
        canary_percent=None,
        rollout_strategy=None,
        auto_complete_rollout=True,
        max_surge_percent=20,
    )
    mock_service_controller.reset_mock()

    runner.invoke(rollout, args=["--no-auto-complete-rollout", "-f", "file"])
    mock_service_controller.rollout.assert_called_once_with(
        SERVICE_CONFIG_DICT,
        name=None,
        version=None,
        canary_percent=None,
        rollout_strategy=None,
        auto_complete_rollout=False,
        max_surge_percent=None,
    )
    mock_service_controller.reset_mock()

    # Provided both -i / --in-place and --rollout-strategy should error.
    result = runner.invoke(
        rollout, args=["-i", "--rollout-strategy=ROLLOUT", "-f", "file"]
    )
    assert result.exception is not None

    result = runner.invoke(
        rollout, args=["--in-place", "--rollout-strategy=ROLLOUT", "-f", "file"]
    )
    assert result.exception is not None


def test_rollback(mock_service_controller: MockServiceController):
    """Tests the rollback command."""

    runner = CliRunner()

    runner.invoke(rollback, args=["--service-id", "test-service-id"])
    mock_service_controller.rollback.assert_called_once_with(
        "test-service-id", None,
    )
    mock_service_controller.reset_mock()

    runner.invoke(
        rollback, args=["--service-id", "test-service-id", "--max-surge-percent", "5"]
    )
    mock_service_controller.rollback.assert_called_once_with(
        "test-service-id", 5,
    )



import inspect
import logging
import os
from sys import path
from typing import Any, Dict, List, Optional

from anyscale import version


logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(os.environ.get("ANYSCALE_LOGLEVEL", "WARN"))

anyscale_dir = os.path.dirname(os.path.abspath(__file__))
path.append(os.path.join(anyscale_dir, "client"))
path.append(os.path.join(anyscale_dir, "sdk"))

import anyscale
from anyscale import compute_config, image, integrations, job, service
from anyscale._private.anyscale_client import AnyscaleClient, AnyscaleClientInterface
from anyscale._private.sdk.base_sdk import Timer
from anyscale.authenticate import AuthenticationBlock
from anyscale.cli_logger import BlockLogger
from anyscale.cluster import get_job_submission_client_cluster_info
from anyscale.cluster_compute import get_cluster_compute_from_name
from anyscale.compute_config import ComputeConfigSDK
from anyscale.connect import ClientBuilder
from anyscale.job import JobSDK
from anyscale.sdk.anyscale_client.sdk import AnyscaleSDK
from anyscale.service import ServiceSDK


# Note: indentation here matches that of connect.py::ClientBuilder.
BUILDER_HELP_FOOTER = """
        See ``anyscale.ClientBuilder`` for full documentation of
        this experimental feature."""

# Auto-add all Anyscale connect builder functions to the top-level.
for attr, _ in inspect.getmembers(ClientBuilder, inspect.isfunction):
    if attr.startswith("_"):
        continue

    def _new_builder(attr: str) -> Any:
        target = getattr(ClientBuilder, attr)

        def new_session_builder(*a: List[Any], **kw: Dict[str, Any]) -> Any:
            builder = ClientBuilder()
            return target(builder, *a, **kw)

        new_session_builder.__name__ = attr
        new_session_builder.__doc__ = target.__doc__ + BUILDER_HELP_FOOTER
        new_session_builder.__signature__ = inspect.signature(target)  # type: ignore

        return new_session_builder

    globals()[attr] = _new_builder(attr)

__version__ = version.__version__

ANYSCALE_ENV = os.environ.copy()


class Anyscale:
    def __init__(
        self, *, auth_token: Optional[str] = None, _host: Optional[str] = None,
    ):
        auth_block = AuthenticationBlock(
            cli_token=auth_token, host=_host, raise_structured_exception=True,
        )
        self._anyscale_client = AnyscaleClient(
            api_clients=(auth_block.anyscale_api_client, auth_block.api_client),
            host=_host,
        )
        self._job_sdk = JobSDK(client=self._anyscale_client)
        self._service_sdk = ServiceSDK(client=self._anyscale_client)
        self._compute_config_sdk = ComputeConfigSDK(client=self._anyscale_client)

    @classmethod
    def _init_private(
        cls, *, client: AnyscaleClientInterface, logger: BlockLogger, timer: Timer
    ):
        # Private constructor used to inject fakes for testing.
        obj = cls.__new__(cls)
        super(Anyscale, obj).__init__()
        obj._anyscale_client = client  # noqa: SLF001
        obj._job_sdk = JobSDK(client=client, logger=logger, timer=timer)  # noqa: SLF001
        obj._service_sdk = ServiceSDK(  # noqa: SLF001
            client=client, logger=logger, timer=timer
        )
        obj._compute_config_sdk = ComputeConfigSDK(  # noqa: SLF001
            client=client, logger=logger, timer=timer
        )
        return obj

    @property
    def job(self) -> JobSDK:  # noqa: F811
        return self._job_sdk

    @property
    def service(self) -> ServiceSDK:  # noqa: F811
        return self._service_sdk

    @property
    def compute_config(self) -> ComputeConfigSDK:  # noqa: F811
        return self._compute_config_sdk



from unittest.mock import Mock, patch

import pytest

from anyscale.controllers.base_controller import BaseController


@pytest.mark.parametrize("initialize_auth_api_client", [True, False])
def test_api_client_authenticated(
    initialize_auth_api_client: bool,
    base_mock_api_client,
    base_mock_anyscale_api_client,
):
    """
    Test a controller that inherits from BaseController has api_client and anyscale_api_client correctly
    populated if initialize_auth_api_client==True and raises an error when these fields
    are called if initialize_auth_api_client==False.
    """

    class Controller(BaseController):
        def test_call_api_client(self,):
            return self.api_client

        def test_call_anyscale_api_client(self,):
            return self.anyscale_api_client

    mock_get_auth_api_client = Mock(
        return_value=Mock(
            api_client=base_mock_api_client,
            anyscale_api_client=base_mock_anyscale_api_client,
        )
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=mock_get_auth_api_client,
    ):
        controller = Controller(initialize_auth_api_client=initialize_auth_api_client)
    if not initialize_auth_api_client:
        mock_get_auth_api_client.assert_not_called()
        with pytest.raises(Exception):  # noqa: PT011
            controller.test_call_api_client()
        with pytest.raises(Exception):  # noqa: PT011
            controller.test_call_anyscale_api_client()
    else:
        # Assert API expections are not structured (and are wrapped as `click.ClickException`) by default
        mock_get_auth_api_client.assert_called_once_with(
            cli_token=None, raise_structured_exception=False
        )
        assert controller.api_client == base_mock_api_client
        assert controller.anyscale_api_client == base_mock_anyscale_api_client



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class ClusterStartup(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'start_time': 'datetime',
        'cluster_id': 'str',
        'id': 'str',
        'finish_time': 'datetime'
    }

    attribute_map = {
        'start_time': 'start_time',
        'cluster_id': 'cluster_id',
        'id': 'id',
        'finish_time': 'finish_time'
    }

    def __init__(self, start_time=None, cluster_id=None, id=None, finish_time=None, local_vars_configuration=None):  # noqa: E501
        """ClusterStartup - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._start_time = None
        self._cluster_id = None
        self._id = None
        self._finish_time = None
        self.discriminator = None

        self.start_time = start_time
        if cluster_id is not None:
            self.cluster_id = cluster_id
        self.id = id
        if finish_time is not None:
            self.finish_time = finish_time

    @property
    def start_time(self):
        """Gets the start_time of this ClusterStartup.  # noqa: E501

        When the startup was initiated. Should be timezone-aware.  # noqa: E501

        :return: The start_time of this ClusterStartup.  # noqa: E501
        :rtype: datetime
        """
        return self._start_time

    @start_time.setter
    def start_time(self, start_time):
        """Sets the start_time of this ClusterStartup.

        When the startup was initiated. Should be timezone-aware.  # noqa: E501

        :param start_time: The start_time of this ClusterStartup.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and start_time is None:  # noqa: E501
            raise ValueError("Invalid value for `start_time`, must not be `None`")  # noqa: E501

        self._start_time = start_time

    @property
    def cluster_id(self):
        """Gets the cluster_id of this ClusterStartup.  # noqa: E501

        Cluster ID from sessions table. None if not known yet.  # noqa: E501

        :return: The cluster_id of this ClusterStartup.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id):
        """Sets the cluster_id of this ClusterStartup.

        Cluster ID from sessions table. None if not known yet.  # noqa: E501

        :param cluster_id: The cluster_id of this ClusterStartup.  # noqa: E501
        :type: str
        """

        self._cluster_id = cluster_id

    @property
    def id(self):
        """Gets the id of this ClusterStartup.  # noqa: E501

        Unique identifier for this startup.  # noqa: E501

        :return: The id of this ClusterStartup.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ClusterStartup.

        Unique identifier for this startup.  # noqa: E501

        :param id: The id of this ClusterStartup.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def finish_time(self):
        """Gets the finish_time of this ClusterStartup.  # noqa: E501

        When this cluster startup has finished; None if it is still in progress.  # noqa: E501

        :return: The finish_time of this ClusterStartup.  # noqa: E501
        :rtype: datetime
        """
        return self._finish_time

    @finish_time.setter
    def finish_time(self, finish_time):
        """Sets the finish_time of this ClusterStartup.

        When this cluster startup has finished; None if it is still in progress.  # noqa: E501

        :param finish_time: The finish_time of this ClusterStartup.  # noqa: E501
        :type: datetime
        """

        self._finish_time = finish_time

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ClusterStartup):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ClusterStartup):
            return True

        return self.to_dict() != other.to_dict()
